"""Table metadata-related classes."""
import itertools
import logging
import threading
from contextlib import contextmanager
from math import ceil
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    cast,
)

from psycopg2.sql import SQL, Composable, Identifier
from tqdm import tqdm

from splitgraph.config import (
    SG_CMD_ASCII,
    SPLITGRAPH_API_SCHEMA,
    SPLITGRAPH_META_SCHEMA,
)
from splitgraph.core.common import Tracer, get_temporary_table_id
from splitgraph.core.fragment_manager import ExtraIndexInfo, get_chunk_groups
from splitgraph.core.indexing.range import quals_to_sql
from splitgraph.core.output import pluralise, truncate_list
from splitgraph.core.sql import select
from splitgraph.core.types import Quals, TableSchema
from splitgraph.engine import ResultShape
from splitgraph.engine.postgres.engine import get_change_key
from splitgraph.exceptions import ObjectIndexingError

if TYPE_CHECKING:
    from splitgraph.core.image import Image
    from splitgraph.core.repository import Repository
    from splitgraph.engine.postgres.engine import PostgresEngine

# Output checkout progress every 5MB to trade off between the latency
# of initializing a batch of object applications and not reporting anything at all
_PROGRESS_EVERY = 5 * 1024 * 1024

# Rough ratio between the size of a PG table and a CStore table (picked empirically)
_CSTORE_COMPRESSION = 2.5


def _generate_select_query(
    engine: "PostgresEngine",
    table: bytes,
    columns: Sequence[str],
    qual_sql: Optional[Composable] = None,
    qual_args: Optional[Tuple] = None,
) -> bytes:
    cur = engine.connection.cursor()

    query = (
        SQL("SELECT ")
        + SQL(",").join(Identifier(c) for c in columns)
        + SQL(" FROM " + table.decode("utf-8"))
        + (SQL(" WHERE ") + qual_sql if qual_args else SQL(""))
    )
    query = cur.mogrify(query, qual_args)

    cur.close()
    return cast(bytes, query)


def _generate_table_names(engine: "PostgresEngine", schema: str, tables: List[str]) -> List[bytes]:
    result = []
    cur = engine.connection.cursor()
    for table in tables:
        result.append(cur.mogrify(SQL("{}.{}").format(Identifier(schema), Identifier(table))))
    cur.close()
    logging.debug("Returning tables %r", tables)
    return result


def _delete_temporary_table(engine: "PostgresEngine", schema: str, table: str) -> None:
    """
    Workaround for Multicorn deadlocks at end_scan time, supposed to be run in a separate thread.
    """
    # Delete the temporary table in an autocommitting connection with a small lock
    # timeout: if we fail to delete it because Multicorn is still holding a lock on it etc,
    # we'd rather leave the temporary table in splitgraph_meta than lock up.
    conn = engine.connection
    conn.autocommit = True
    try:
        with conn.cursor() as cur:
            cur.execute("SET lock_timeout TO '5s';")
            cur.execute(
                SQL("DROP TABLE IF EXISTS {}.{}").format(Identifier(schema), Identifier(table))
            )
        logging.info("Dropped temporary table %s", table)
    finally:
        conn.autocommit = False
        conn.close()


def _empty_callback(**kwargs) -> None:
    pass


class QueryPlan:
    """
    Represents the initial query plan (fragments to query) for given columns and
    qualifiers.
    """

    def __init__(self, table: "Table", quals: Optional[Quals], columns: Sequence[str]) -> None:
        self.table = table
        self.quals = quals
        self.columns = columns
        self.tracer = Tracer()

        self.object_manager = table.repository.objects

        # Make sure objects don't repeat (the second copy of the same object does nothing).
        # Some process makes certain datasets have duplicate objects, disabling the singleton
        # "fast path" and requiring materialization.
        self.required_objects = list(dict.fromkeys(table.objects))
        self.tracer.log("resolve_objects")
        self.filtered_objects = self.object_manager.filter_fragments(
            self.required_objects, table, quals
        )
        # Estimate the number of rows in the filtered objects
        object_meta = self.object_manager.get_object_meta(self.filtered_objects)
        self.estimated_rows = sum([o.rows_inserted - o.rows_deleted for o in object_meta.values()])

        # Estimate the row size in bytes by looking at the total object size
        # This can sometimes be important if we're querying wide tables with JSON data where
        # a single cell can be a few KB -- in some cases, this can make PG think it's going to get
        # less data than there actually be and pick suboptimal plans.
        total_rows_inserted = sum(o.rows_inserted for o in object_meta.values())

        self.size_per_row = (
            (
                sum(o.size for o in object_meta.values())
                / sum(o.rows_inserted for o in object_meta.values())
                * _CSTORE_COMPRESSION
            )
            if total_rows_inserted
            else 0
        )

        self.tracer.log("filter_objects")

        # Prepare a list of objects to query

        # Special fast case: single-chunk groups can all be batched together
        # and queried directly without having to copy them to a staging table.
        # We also grab all fragments from multiple-fragment groups and batch them together
        # for future application.
        #
        # Technically, we could do multiple batches of application for these groups
        # (apply first batch to the staging table, extract the result, clean the table,
        # apply next batch etc): in the middle of it we could also talk back to the object
        # manager and release the objects that we don't need so that they can be garbage
        # collected. The tradeoff is that we perform more calls to apply_fragments (hence
        # more roundtrips).
        self.non_singletons, self.singletons = self._extract_singleton_fragments()

        logging.info(
            "Fragment grouping: %d singletons, %d non-singletons",
            len(self.singletons),
            len(self.non_singletons),
        )
        self.tracer.log("group_fragments")

        self.sql_quals, self.sql_qual_vals = quals_to_sql(
            quals, column_types={c.name: c.pg_type for c in self.table.table_schema}
        )

        if self.singletons:
            self.singleton_queries = _generate_table_names(
                self.object_manager.object_engine, SPLITGRAPH_META_SCHEMA, self.singletons
            )
        else:
            self.singleton_queries = []
        self.tracer.log("generate_singleton_queries")

    def _extract_singleton_fragments(self) -> Tuple[List[str], List[str]]:
        # Get fragment boundaries (min-max PKs of every fragment).
        table_pk = get_change_key(self.table.table_schema)
        object_pks = self.object_manager.get_min_max_pks(self.filtered_objects, table_pk)

        surrogate_pk = not any(t.is_pk for t in self.table.table_schema)
        if surrogate_pk:
            object_pks = self.table.repository.objects.generate_surrogate_pk(self.table, object_pks)

        # Group fragments into non-overlapping groups: those can be applied independently of each other.
        object_groups = get_chunk_groups(
            [
                (object_id, min_max[0], min_max[1])
                for object_id, min_max in zip(self.filtered_objects, object_pks)
            ]
        )
        singletons: List[str] = []
        non_singletons: List[str] = []
        for group in object_groups:
            if len(group) == 1:
                singletons.append(group[0][0])
            else:
                non_singletons.extend(object_id for object_id, _, _ in group)
        return non_singletons, singletons


QueryPlanCacheKey = Tuple[Optional[Tuple[Tuple[Tuple[str, str, Any]]]], Tuple[str]]


def _get_plan_cache_key(quals: Optional[Quals], columns: Sequence[str]) -> QueryPlanCacheKey:
    quals = (
        cast(
            Tuple[Tuple[Tuple[str, str, Any]]],
            tuple(tuple(tuple(q) for q in qual) for qual in quals),
        )
        if quals
        else None
    )
    columns = cast(Tuple[str], tuple(columns))
    return quals, columns


def merge_index_data(current_index: Dict[str, Any], new_index: Dict[str, Any]):
    for index_type, index_data in new_index.items():
        for col_name, col_index_data in index_data.items():
            if index_type not in current_index:
                current_index[index_type] = {}
            if col_name not in current_index[index_type]:
                current_index[index_type][col_name] = {}

            current_index[index_type][col_name] = col_index_data


class Table:
    """Represents a Splitgraph table in a given image. Shouldn't be created directly, use Table-loading
    methods in the :class:`splitgraph.core.image.Image` class instead."""

    def __init__(
        self,
        repository: "Repository",
        image: "Image",
        table_name: str,
        table_schema: TableSchema,
        objects: List[str],
    ) -> None:
        self.repository = repository
        self.image = image
        self.table_name = table_name
        self.table_schema = table_schema

        # List of fragments this table is composed of
        self.objects = objects

        self._query_plans: Dict[Tuple[Optional[Quals], Tuple[str]], QueryPlan] = {}

    def __repr__(self) -> str:
        return "Table %s in %s" % (self.table_name, str(self.image))

    def get_query_plan(
        self, quals: Optional[Quals], columns: Sequence[str], use_cache: bool = True
    ) -> QueryPlan:
        """
        Start planning a query (preliminary steps before object downloading,
        like qualifier filtering).

        :param quals: Qualifiers in CNF form
        :param columns: List of columns
        :param use_cache: If True, will fetch the plan from the cache for the same qualifiers and columns.
        :return: QueryPlan
        """
        key = _get_plan_cache_key(quals, columns)

        if use_cache and key in self._query_plans:
            plan = self._query_plans[key]
            # Reset the tracer in the plan: if this instance
            # persisted, the resolve/filter times in it will be
            # way in the past.
            plan.tracer = Tracer()
            plan.tracer.log("resolve_objects")
            plan.tracer.log("filter_objects")
            plan.tracer.log("group_fragments")
            plan.tracer.log("generate_singleton_queries")
            return plan

        plan = QueryPlan(self, quals, columns)
        self._query_plans[key] = plan
        return plan

    def materialize(
        self,
        destination: str,
        destination_schema: Optional[str] = None,
        lq_server: Optional[str] = None,
        temporary: bool = False,
    ) -> None:
        """
        Materializes a Splitgraph table in the target schema as a normal Postgres table, potentially downloading all
        required objects and using them to reconstruct the table.

        :param destination: Name of the destination table.
        :param destination_schema: Name of the destination schema.
        :param lq_server: If set, sets up a layered querying FDW for the table instead using this foreign server.
        """
        # Circular import
        from splitgraph.hooks.data_source.fdw import create_foreign_table

        destination_schema = destination_schema or self.repository.to_schema()
        engine = self.repository.object_engine
        object_manager = self.repository.objects
        engine.delete_table(destination_schema, destination)

        if not lq_server:
            # Materialize by applying fragments to one another in their dependency order.
            with object_manager.ensure_objects(
                table=self, objects=self.objects
            ) as required_objects:
                engine.create_table(
                    schema=destination_schema,
                    table=destination,
                    schema_spec=self.table_schema,
                    include_comments=True,
                    unlogged=True,
                    temporary=temporary,
                )
                if required_objects:
                    logging.debug("Applying %s...", pluralise("fragment", len(required_objects)))

                    table_size = self.get_size()

                    progress_every: Optional[int]
                    if table_size > _PROGRESS_EVERY:
                        progress_every = int(
                            ceil(len(required_objects) * _PROGRESS_EVERY / float(table_size))
                        )
                    else:
                        progress_every = None

                    engine.apply_fragments(
                        [(SPLITGRAPH_META_SCHEMA, d) for d in cast(List[str], required_objects)],
                        destination_schema,
                        destination,
                        progress_every=progress_every,
                    )
        else:
            query, args = create_foreign_table(
                destination_schema,
                lq_server,
                self.table_name,
                self.table_schema,
                extra_options={"table": self.table_name},
            )

            engine.run_sql(query, args)

    def query_indirect(
        self, columns: List[str], quals: Optional[Quals]
    ) -> Tuple[Iterator[bytes], Callable, QueryPlan]:
        """
        Run a read-only query against this table without materializing it. Instead of
        actual results, this returns a generator of SQL queries that the caller can use
        to get the results as well as a callback that the caller has to run after they're
        done consuming the results.

        In particular, the query generator will prefer returning direct queries to
        Splitgraph objects and only when those are exhausted will it start materializing
        delta-compressed fragments.

        This is an advanced method: you probably want to call table.query().

        :param columns: List of columns from this table to fetch
        :param quals: List of qualifiers in conjunctive normal form. See the documentation for
            FragmentManager.filter_fragments for the actual format.
        :return: Generator of queries (bytes), a callback and a query plan object (containing stats
            that are fully populated after the callback has been called to end the query).
        """
        plan = self.get_query_plan(quals, columns)
        required_objects = plan.filtered_objects
        logging.info(
            "Using %d fragments (%s) to satisfy the query",
            len(required_objects),
            truncate_list(required_objects),
        )
        if not required_objects:
            return cast(Iterator[bytes], []), cast(Callable, _empty_callback), plan

        object_manager = self.repository.objects
        if not plan.non_singletons:
            with object_manager.ensure_objects(
                self, objects=required_objects, defer_release=True, tracer=plan.tracer
            ) as eo_result:
                _, release_callback = cast(Tuple, eo_result)
                return iter(plan.singleton_queries), cast(Callable, release_callback), plan

        def _generate_nonsingleton_query() -> Generator[bytes, None, None]:
            # If we have fragments that need applying to a staging area, we don't want to
            # do it immediately: the caller might be satisfied with the data they got from
            # the queries to singleton fragments. So here we have a callback that, when called,
            # actually materializes the chunks into a temporary table and then changes
            # the table's release callback to also delete that temporary table.

            # There's a slight issue: we can't use temporary tables if we're returning
            # pointers to tables since the caller might be in a different session.
            staging_table = self._create_staging_table()
            engine = self.repository.object_engine

            def _f(from_fdw=False):
                # This is very horrible but the way we share responsibilities between ourselves
                # and Multicorn leaves us no other choice. In LQ, this is supposed to be called
                # during EndForeignScan (at which point we are done with this staging table).
                # However, EndForeignScan doesn't actually release locks on tables that Multicorn
                # was reading (that are acquired outside of our control if the foreign scan happened in a
                # transaction). In that case we can't delete this table until the transaction actually finishes
                # -- which can't happen until we've deleted this table.

                # Other options are: generating the materialization query and running it on
                # Multicorn side (issues with Portals not being able to perform DDL and us having
                # to rework our interface), adding a DROP table to the end of the query (then
                # it doesn't return results, Portals still can't do DDL and the query is no longer
                # idempotent).

                # Instead, we pretend that we've successfully cleaned up but actually spawn
                # a thread whose single job will be running DROP table and waiting until it actually
                # returns.

                if from_fdw:
                    thread = threading.Thread(
                        target=_delete_temporary_table,
                        args=(engine, SPLITGRAPH_META_SCHEMA, staging_table),
                    )
                    thread.start()
                else:
                    engine.delete_table(SPLITGRAPH_META_SCHEMA, staging_table)

            nonlocal release_callback
            release_callback.append(_f)

            # Apply the fragments (just the parts that match the qualifiers) to the staging area
            if quals:
                engine.apply_fragments(
                    [(SPLITGRAPH_META_SCHEMA, o) for o in plan.non_singletons],
                    SPLITGRAPH_META_SCHEMA,
                    staging_table,
                    extra_quals=plan.sql_quals,
                    extra_qual_args=plan.sql_qual_vals,
                    schema_spec=self.table_schema,
                )
            else:
                engine.apply_fragments(
                    [(SPLITGRAPH_META_SCHEMA, o) for o in plan.non_singletons],
                    SPLITGRAPH_META_SCHEMA,
                    staging_table,
                    schema_spec=self.table_schema,
                )
            engine.commit()
            table_name = _generate_table_names(engine, SPLITGRAPH_META_SCHEMA, [staging_table])[0]
            yield table_name

        with object_manager.ensure_objects(
            self, objects=required_objects, defer_release=True, tracer=plan.tracer
        ) as eo_result:
            _, release_callback = cast(Tuple, eo_result)
            return (
                itertools.chain(plan.singleton_queries, _generate_nonsingleton_query()),
                cast(Callable, release_callback),
                plan,
            )

    @contextmanager
    def query_lazy(self, columns: List[str], quals: Quals) -> Iterator[Iterator[Dict[str, Any]]]:
        """
        Run a read-only query against this table without materializing it.

        :param columns: List of columns from this table to fetch
        :param quals: List of qualifiers in conjunctive normal form. See the documentation for
            FragmentManager.filter_fragments for the actual format.
        :return: Generator of dictionaries of results.
        """

        table_gen, release_callback, plan = self.query_indirect(columns, quals)
        engine = self.repository.object_engine

        def _generate_results():
            for table in table_gen:
                result = engine.run_sql(
                    _generate_select_query(
                        engine, table, plan.columns, plan.sql_quals, plan.sql_qual_vals
                    )
                )
                for row in result:
                    yield {c: v for c, v in zip(columns, row)}

        try:
            yield _generate_results()
        finally:
            release_callback()

    def query(self, columns: List[str], quals: Quals):
        """
        Run a read-only query against this table without materializing it.

        This is a wrapper around query_lazy() that force evaluates the results which
        might mean more fragments being materialized that aren't needed.

        :param columns: List of columns from this table to fetch
        :param quals: List of qualifiers in conjunctive normal form. See the documentation for
            FragmentManager.filter_fragments for the actual format.
        :return: List of dictionaries of results
        """
        with self.query_lazy(columns, quals) as result:
            return list(result)

    def get_size(self) -> int:
        """
        Get the physical size used by the table's objects (including those shared with other tables).

        This is calculated from the metadata, the on-disk footprint might be smaller if not all of table's
        objects have been downloaded.

        :return: Size of the table in bytes.
        """
        return cast(
            int,
            self.repository.engine.run_sql(
                select("get_table_size", table_args="(%s,%s,%s,%s)", schema=SPLITGRAPH_API_SCHEMA),
                (
                    self.repository.namespace,
                    self.repository.repository,
                    self.image.image_hash,
                    self.table_name,
                ),
                return_shape=ResultShape.ONE_ONE,
            )
            or 0,
        )

    def get_length(self) -> int:
        """
        Get the number of rows in this table.

        This might be smaller than the total number of rows in all objects belonging to this
        table as some objects might overwrite each other.

        :return: Number of rows in table
        """
        return cast(
            int,
            self.repository.engine.run_sql(
                select(
                    "get_table_length", table_args="(%s,%s,%s,%s)", schema=SPLITGRAPH_API_SCHEMA
                ),
                (
                    self.repository.namespace,
                    self.repository.repository,
                    self.image.image_hash,
                    self.table_name,
                ),
                return_shape=ResultShape.ONE_ONE,
            )
            or 0,
        )

    def reindex(self, extra_indexes: ExtraIndexInfo, raise_on_patch_objects=True) -> List[str]:
        """
        Run extra indexes on all objects in this table and update their metadata.
        This only works on objects that don't have any deletions or upserts (have a deletion hash of 000000...).

        :param extra_indexes: Dictionary of {index_type: column: index_specific_kwargs}.
        :param raise_on_patch_objects: If True, will raise an exception if any objects in the table
            overwrite any other objects. If False, will log a warning but will reindex all non-patch objects.

        :returns List of objects that were reindexed.
        """
        object_manager = self.repository.objects

        object_meta = object_manager.get_object_meta(self.objects)

        # Because the index includes both inserted and deleted entries (so that we know
        # to grab the object if it deletes something we're interested in), we can't easily
        # reindex objects that include deletions since we'd have to find the object with
        # the old value for whatever we're deleting and also download it.

        valid_objects = {o: om for o, om in object_meta.items() if om.deletion_hash == "0" * 64}
        if len(valid_objects) < len(object_meta):
            message = "%d object(s) in %s have deletions or upserts, reindexing unsupported" % (
                len(object_meta) - len(valid_objects),
                self,
            )
            if raise_on_patch_objects:
                raise ObjectIndexingError(message)
            else:
                logging.warning(message)

        # Download the objects to reindex them
        with object_manager.ensure_objects(self, objects=list(valid_objects)):
            for object_id in tqdm(valid_objects, unit="objs", ascii=SG_CMD_ASCII):
                current_index = valid_objects[object_id].object_index

                index_struct = object_manager.generate_object_index(
                    object_id, self.table_schema, changeset=None, extra_indexes=extra_indexes
                )

                # Update the index in-place and overwrite it
                merge_index_data(current_index, index_struct)

        object_manager.register_objects(list(valid_objects.values()))
        return list(valid_objects)

    def _create_staging_table(self) -> str:
        staging_table = get_temporary_table_id()

        logging.debug("Using staging table %s", staging_table)
        self.repository.object_engine.create_table(
            schema=SPLITGRAPH_META_SCHEMA,
            table=staging_table,
            schema_spec=self.table_schema,
            unlogged=True,
        )
        return staging_table



""" The CONFIG object is created and exported once __at import time__
    Calling CONFIG["KEY"] directly should be sufficient in most cases,
    except when a config value has changed since importing CONFIG.
    In that case, create_config_dict() can provide an updated config dict


    Priority (highest to lowest):

        1. Command line argument values
        2. Environment variable values
        3. Config file values in [defaults] section
        4. DEFAULTS (see keys.py)

"""
from .config import create_config_dict, get_singleton

CONFIG = create_config_dict()

PG_HOST = get_singleton(CONFIG, "SG_ENGINE_HOST")
PG_PORT = get_singleton(CONFIG, "SG_ENGINE_PORT")
PG_DB = get_singleton(CONFIG, "SG_ENGINE_DB_NAME")
PG_USER = get_singleton(CONFIG, "SG_ENGINE_USER")
PG_PWD = get_singleton(CONFIG, "SG_ENGINE_PWD")

SPLITGRAPH_META_SCHEMA = get_singleton(CONFIG, "SG_META_SCHEMA")
SPLITGRAPH_API_SCHEMA = "splitgraph_api"

FDW_CLASS = get_singleton(CONFIG, "SG_FDW_CLASS")

SG_CMD_ASCII = get_singleton(CONFIG, "SG_CMD_ASCII") == "true"

DEFAULT_CHUNK_SIZE = int(get_singleton(CONFIG, "SG_COMMIT_CHUNK_SIZE"))

REMOTES = list(CONFIG.get("remotes", []))

# This is a global variable that gets flipped to True by the Multicorn FDW class
# at startup. When we're running from within an engine as an FDW, we might need to use
# different connection parameters to connect to other engines. It's not trivial to detect
# whether we're running inside of an embedded Python otherwise and this variable needs to
# ultimately make it into all get_engine() constructors, so this is simpler than threading
# it through all calls that FDW makes.
IN_FDW = False



"""
Utilities to generate a sample dbt project that works on Splitgraph-loaded data
"""
import os
from pathlib import Path
from typing import Any, Dict, List, Tuple

import ruamel.yaml

from splitgraph.cloud.project.templates import (
    DBT_PROJECT_TEMPLATE,
    SOURCE_TEMPLATE_NOCOM,
    SOURCES_YML_TEMPLATE,
)
from splitgraph.cloud.project.utils import get_source_name
from splitgraph.core.repository import Repository
from splitgraph.utils.yaml import safe_load


def generate_dbt_project(repositories: List[str], basedir: Path) -> None:
    """
    Generate a sample dbt project in a directory
    :param repositories: List of repository names used by this dbt project
    :param basedir: Directory to put the project in
    """

    # Generate dbt_project.yml
    with open(os.path.join(basedir, "dbt_project.yml"), "w") as f:
        f.write(DBT_PROJECT_TEMPLATE)

    # Generate models/staging/sources.yml
    yml = ruamel.yaml.YAML()
    sources_yml = yml.load(SOURCES_YML_TEMPLATE)
    sources_yml["sources"][0]["name"] = get_source_name(repositories[0])
    sources_yml["sources"][0]["schema"] = repositories[0]

    # Add the remaining sources without comments
    for repository in repositories[1:]:
        sources_yml_nocom = safe_load(SOURCE_TEMPLATE_NOCOM)
        sources_yml_nocom["name"] = get_source_name(repository)
        sources_yml_nocom["schema"] = repository
        sources_yml["sources"].append(sources_yml_nocom)

    os.makedirs(os.path.join(basedir, "models/staging"), exist_ok=True)
    with open(os.path.join(basedir, "models/staging/sources.yml"), "w") as f:
        yml.dump(sources_yml, f)

    # Generate models/staging/[source_name]/[source_name.sql]
    for repository in repositories:
        source_name = get_source_name(repository)
        model_path = os.path.join(basedir, "models/staging", source_name)
        os.makedirs(model_path, exist_ok=True)
        with open(os.path.join(model_path, source_name + ".sql"), "w") as f:
            f.write(  # nosec
                f"""SELECT 
  *
FROM {{{{ source('{source_name.replace("'", "''")}', 'some_table') }}}}
"""
            )


def _make_source(repository: str) -> Dict[str, str]:
    repo_obj = Repository.from_schema(repository)
    return {
        "dbt_source_name": get_source_name(repository),
        "namespace": repo_obj.namespace,
        "repository": repo_obj.repository,
        "hash_or_tag": "latest",
    }


def generate_dbt_plugin_params(repositories: List[str]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Generate the required configuration parameters for the Splitgraph dbt plugin"""

    # At runtime (when we execute the GitHub action), if we're checking the dbt project
    # into the same repository where splitgraph.yml lives (they can coexist), we need a way
    # to send the repo over to Splitgraph Cloud. We put a placeholder here instead and construct
    # the Git pull URL at action runtime (using GITHUB_TOKEN).
    credentials = {"git_url": "$THIS_REPO_URL"}

    # Same with the branch: we want to inject the current SHA we're running the action for.
    params = {"sources": [_make_source(r) for r in repositories], "git_branch": "$THIS_SHA"}

    return params, credentials



"""
sgr commands related to building and rebuilding Splitfiles.
"""
import os

import click

from .common import ImageType, RepositoryType


@click.command(name="build")
@click.argument("splitfile", type=click.File("r"))
@click.option(
    "-a",
    "--args",
    multiple=True,
    type=(str, str),
    help="Parameters to be substituted into the Splitfile. All parameters mentioned in the file"
    " must be specified in order for the Splitfile to be executed.",
)
@click.option(
    "-o", "--output-repository", help="Repository to store the result in.", type=RepositoryType()
)
def build_c(splitfile, args, output_repository):
    """
    Build Splitgraph images.

    This executes a Splitfile, building a new image or checking it out from cache if the same
    image had already been built.

    Examples:

    ``sgr build my.splitfile``

    Executes ``my.splitfile`` and writes its output into a new repository with the same name
    as the Splitfile (my) unless the name is specified in the Splitfile.

    ``sgr build my.splitfile -o mynew/repo``

    Executes ``my.splitfile`` and writes its output into ``mynew/repo``.

    ``sgr build my_other.splitfile -o mynew/otherrepo --args PARAM1 VAL1 --args PARAM2 VAL2``

    Executes ``my_other.splitfile`` with parameters ``PARAM1`` and ``PARAM2`` set to
    ``VAL1`` and  ``VAL2``, respectively.
    """
    from splitgraph.core.repository import Repository
    from splitgraph.splitfile.execution import execute_commands

    args = {k: v for k, v in args}
    click.echo("Executing Splitfile %s with arguments %r" % (splitfile.name, args))

    if output_repository is None:
        file_name = os.path.splitext(os.path.basename(splitfile.name))[0]
        output_repository = Repository.from_schema(file_name)

    execute_commands(splitfile.read(), args, output=output_repository)


@click.command(name="provenance")
@click.argument("image_spec", type=ImageType(get_image=True))
@click.option(
    "-f",
    "--full",
    required=False,
    is_flag=True,
    help="Recreate the Splitfile used to create this image",
)
@click.option(
    "-e",
    "--ignore-errors",
    is_flag=True,
    help="If set, ignore commands that aren't reproducible (like MOUNT or custom commands)",
)
def provenance_c(image_spec, full, ignore_errors):
    """
    Reconstruct the provenance of an image.

    This inspects the image to produce a list of images that were used by the Splitfile
    that created it, or a Splitfile with the same effect.

    `IMAGE_SPEC` must be of the form `[NAMESPACE/]REPOSITORY[:HASH_OR_TAG]`.
    If no tag is specified, `latest` is used.

    Examples:

    Assume `my/repo` is produced by the following Splitfile:

    ```
    FROM MOUNT [...] IMPORT external_table
    FROM noaa/climate IMPORT {SELECT * FROM rainfall_data WHERE state = 'AZ'} AS rainfall_data
    ```

    `my/repo` will have 2 images: one having `hash_1` (with the `external_table` imported from a mounted database)
    and one having `hash_2` (with both `external_table` and the `rainfall_data` containing the result
    of the query run against the then-latest image in the `noaa/climate` repository).

    In this case:

    ```
    sgr provenance my/repo
    ```

    Returns a list of repositories and images that were imported by the Splitfile that constructed this image::

    ```
    my/repo:[hash_2] depends on:
    noaa/climate:[hash_3]
    ```

    Where `hash_3` is the hash of the latest image in the `noaa/climate` repository at the time the original
    Splitfile was run. However:

    ```
    sgr provenance -f my/repo
    ```

    Will try to reconstruct the Splitfile that can be used to build this image. Since the FROM MOUNT command isn't
    reproducible (requires access to the original external database, which is a moving target), this will fail.

    If `-e` is passed, this will emit information about irreproducible commands instead of failing.

    ```
    sgr provenance -ef my/repo

    # Splitfile commands used to reconstruct my/repo:[image_hash]
    # Irreproducible Splitfile command of type MOUNT
    FROM noaa/climate:[hash_3] IMPORT {SELECT * FROM rainfall_data WHERE state = 'AZ'}
    ```
    """
    repository, image = image_spec

    if full:
        splitfile_commands = image.to_splitfile(ignore_irreproducible=ignore_errors)
        click.echo(
            "# Splitfile commands used to recreate %s:%s" % (str(repository), image.image_hash)
        )
        click.echo("\n".join(splitfile_commands))
    else:
        result = image.provenance()
        click.echo("%s:%s depends on:" % (str(repository), image.image_hash))
        click.echo("\n".join("%s:%s" % rs for rs in result))


@click.command(name="dependents")
@click.argument("image_spec", type=ImageType(get_image=False))
@click.option(
    "-O",
    "--source-on",
    type=str,
    default=None,
    help="Override the engine to look the source up on",
)
@click.option(
    "-o",
    "--dependents-on",
    type=str,
    default=None,
    help="Override the engine to list dependents from",
)
def dependents_c(image_spec, source_on, dependents_on):
    """
    List images that were created from an image.

    This is the inverse of the sgr provenance command. It will list all images that were
    created using a Splitfile that imported data from this image.

    By default, this will look at images on the local engine. The engine can be overridden
    with --source-on and --dependents-on. For example:

        sgr dependents --source-on data.splitgraph.com --dependents-on LOCAL noaa/climate:latest

    will show all images on the local engine that derived data from `noaa/climate:latest`
    on the Splitgraph registry.
    """
    from splitgraph.core.repository import Repository
    from splitgraph.engine import get_engine

    source_engine = get_engine(source_on) if source_on else get_engine()
    repository, image = image_spec
    repository = Repository.from_template(repository, engine=source_engine)
    image = repository.images[image]

    target_engine = get_engine(dependents_on) if dependents_on else get_engine()

    result = image.provenance(reverse=True, engine=target_engine)
    click.echo("%s:%s is depended on by:" % (str(repository), image.image_hash))
    click.echo("\n".join("%s:%s" % rs for rs in result))


@click.command(name="rebuild")
@click.argument("image_spec", type=ImageType(get_image=True))
@click.option(
    "-u",
    "--update",
    is_flag=True,
    help="Rederive the image against the latest version of all dependencies.",
)
@click.option(
    "-a",
    "--against",
    multiple=True,
    type=ImageType(),
    help="Images to substitute into the reconstructed Splitfile, of the form"
    " [NAMESPACE/]REPOSITORY[:HASH_OR_TAG]. Default tag is 'latest'.",
)
def rebuild_c(image_spec, update, against):
    """
    Rebuild images against different dependencies.

    Examines the provenance of a Splitgraph image created by a Splitfile and reruns it against different images than
    the ones that were imported by the original run.

    Examples:

    ``sgr rebuild my/repo --against noaa/climate:old_data``

    Reconstructs the Splitfile used to create ``my/repo:latest``, replaces all imports from ``noaa/climate`` with
    imports from ``noaa/climate:old_data`` and reruns the Splitfile.

    ``sgr rebuild my/repo:other_tag -u``

    Rebuilds ``my_repo:other_tag`` against the latest versions of all of its dependencies.

    Image caching still works in this case: if the result of the rebuild already exists, the image will be checked
    out.
    """
    repository, image = image_spec

    # Replace the sources used to construct the image with either the latest ones or the images specified by the user.
    # This doesn't require us at this point to have pulled all the dependencies: the Splitfile executor will do it
    # after we feed in the reconstructed and patched Splitfile.
    deps = {k: v for k, v in image.provenance()}
    new_images = (
        {repo: repl_image for repo, repl_image in against}
        if not update
        else {repo: "latest" for repo, _ in deps.items()}
    )
    deps.update(new_images)

    click.echo("Rerunning %s:%s against:" % (str(repository), image.image_hash))
    click.echo("\n".join("%s:%s" % rs for rs in new_images.items()))

    from splitgraph.splitfile.execution import rebuild_image

    rebuild_image(image, new_images)



from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from splitgraph.core.image_mounting import ImageMounter
from splitgraph.core.types import (
    Credentials,
    IntrospectionResult,
    Params,
    TableInfo,
    TableParams,
    get_table_list,
)
from splitgraph.hooks.data_source.base import LoadableDataSource, TransformingDataSource
from splitgraph.ingestion.dbt.utils import (
    compile_dbt_manifest,
    run_dbt_transformation_from_git,
)

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine


class DBTDataSource(TransformingDataSource, LoadableDataSource):

    table_params_schema = {"type": "object", "properties": {}}

    params_schema = {
        "type": "object",
        "properties": {
            "git_branch": {
                "type": "string",
                "title": "Git branch",
                "description": "Branch or commit hash to use for the dbt project.",
                "default": "master",
            },
            "sources": {
                "type": "array",
                "title": "dbt source map",
                "description": "Describe which Splitgraph image to use for each dbt source "
                "(defined in sources.yml).",
                "items": {
                    "type": "object",
                    "properties": {
                        "dbt_source_name": {
                            "type": "string",
                            "title": "Source name",
                            "description": "Source name in the dbt model",
                        },
                        "namespace": {
                            "type": "string",
                            "title": "Namespace",
                        },
                        "repository": {
                            "type": "string",
                            "title": "Repository",
                        },
                        "hash_or_tag": {
                            "type": "string",
                            "title": "Hash or tag",
                            "default": "latest",
                        },
                    },
                    "required": ["dbt_source_name", "namespace", "repository"],
                },
            },
        },
    }

    credentials_schema = {
        "type": "object",
        "properties": {
            # This is a secret since the git URL might have a password in it.
            "git_url": {
                "type": "string",
                "title": "dbt model Git URL",
                "description": "URL to the Git repo "
                "with the dbt project, for example,"
                "`https://uname:pass_or_token@github.com/organisation/repository.git`.",
            },
        },
        "required": ["git_url"],
    }

    _icon_file = "dbt.svg"

    def __init__(
        self,
        engine: "PostgresEngine",
        credentials: Credentials,
        params: Params,
        tables: Optional[TableInfo] = None,
        image_mounter: Optional[ImageMounter] = None,
    ):
        super().__init__(engine, credentials, params, tables, image_mounter)

        self.git_url: str = self.credentials["git_url"]
        self.git_branch: str = self.params.get("git_branch", "master")
        self.source_map: Dict[str, Tuple[str, str, str]] = {
            source["dbt_source_name"]: (
                source["namespace"],
                source["repository"],
                source.get("hash_or_tag", "latest"),
            )
            for source in self.params.get("sources", [])
        }

    @classmethod
    def get_name(cls) -> str:
        return "dbt"

    @classmethod
    def get_description(cls) -> str:
        return "Create a Splitgraph repository from a dbt model"

    def get_required_images(self) -> List[Tuple[str, str, str]]:
        return sorted(set(self.source_map.values()))

    def introspect(self) -> IntrospectionResult:
        # Get the dbt manifest file
        manifest = compile_dbt_manifest(
            engine=self.engine, repository_url=self.git_url, repository_ref=self.git_branch
        )

        # Extract all models from the repository that materialize as "table" and aren't tests
        model_names = [
            n["name"]
            for n in manifest["nodes"].values()
            if n["config"]["materialized"] == "table" and n["resource_type"] == "model"
        ]
        return IntrospectionResult({m: ([], TableParams({})) for m in model_names})

    def _load(self, schema: str, tables: Optional[TableInfo] = None):
        with self.mount_required_images() as schema_map:
            # Build a map of dbt data sources to schemas
            dbt_source_map = {
                source_name: schema_map[image] for source_name, image in self.source_map.items()
            }

            # Get a list of models to run, based on the tables we were passed
            models = get_table_list(tables) if tables else None

            # Run dbt against our source schemas and write data out into the target schema
            run_dbt_transformation_from_git(
                self.engine,
                schema,
                self.git_url,
                self.git_branch,
                source_schema_map=dbt_source_map,
                models=models,
            )



import os
from typing import Optional


def get_environment_config_value(key: str, default_return: Optional[str] = None) -> Optional[str]:
    """Get the environment variable value of the environment variable matching key.
    Otherwise return default_return.
    """

    return os.environ.get(key, default_return)



"""Command line tools for ingesting/exporting Splitgraph images into other formats."""
import io
import logging
from itertools import islice

import click

from splitgraph.commandline.common import ImageType, RepositoryType
from splitgraph.core.output import ResettableStream


@click.group(name="csv")
def csv_group():
    """Import/export Splitgraph images in CSV format."""


@click.command(name="export")
@click.argument("image_spec", type=ImageType(default=None))
@click.argument("query")
@click.option(
    "-f",
    "--file",
    type=click.File("w"),
    default="-",
    help="File name to export to, default stdout.",
)
@click.option(
    "-l",
    "--layered",
    help="Don't materialize the tables, use layered querying instead.",
    is_flag=True,
    default=False,
)
def csv_export(image_spec, query, file, layered):
    """
    Export the result of a query as CSV.

    Examples:

    `sgr csv export noaa/climate "SELECT * FROM rainfall"`

    Output everything in the currently checked-out `"rainfall"` table as CSV.

    `sgr csv export noaa/climate:dec_2018 "SELECT * FROM rainfall WHERE state = 'AZ' -f dec_2018_az.csv`

    Check out the `dec_2018` tag of `noaa/climate` and output values from `"rainfall"` for Arizona to `dec_2018_az.csv`

    `sgr csv export --layered noaa/climate:abcdef1234567890 "SELECT * FROM rainfall JOIN other_table ON..."`

    Uses layered querying instead to execute a join on tables in a certain image (satisfying the query without
    having to check the image out).
    """
    from splitgraph.ingestion.csv import csv_adapter

    repository, image = image_spec
    csv_adapter.to_data(query, image, repository, layered, buffer=file)


@click.command(name="import")
@click.argument("repository", type=RepositoryType(exists=True))
@click.argument("table")
@click.option(
    "-f",
    "--file",
    type=click.File("rb"),
    default="-",
    help="File name to import data from, default stdin.",
)
@click.option(
    "-r", "--replace", default=False, is_flag=True, help="Replace the table if it already exists."
)
@click.option(
    "-k",
    "--primary-key",
    multiple=True,
    help="Use the specified column(s) as primary key(s)",
    default=[],
)
@click.option(
    "-t",
    "--override-type",
    multiple=True,
    type=(str, str),
    help="Explicitly set types of these columns to PG types",
)
@click.option("--encoding", help="Encoding to use for the CSV file", default=None)
@click.option("--separator", default=",", help="CSV separator to use")
@click.option(
    "--no-header",
    default=False,
    is_flag=True,
    help="Treats the first line of the CSV as data rather than a header.",
)
@click.option(
    "--skip-schema-check",
    default=False,
    is_flag=True,
    help="Skips checking that the dataframe is compatible with the target schema.",
)
@click.option(
    "--ignore-empty-strings",
    is_flag=True,
    help="When doing schema inference, treat empty strings as NULLs",
)
def csv_import(
    repository,
    table,
    file,
    replace,
    primary_key,
    override_type,
    encoding,
    separator,
    no_header,
    skip_schema_check,
    ignore_empty_strings,
):
    """
    Import a CSV file into a checked-out Splitgraph repository. This doesn't create a new image, use `sgr commit`
    after the import and any adjustments (e.g. adding primary keys or converting column types) to do so.

    If the target table doesn't exist, this will create a new table.

    If the target table does exist, this will try and patch the new values in by updating rows that exist in the
    current table (as per its primary key constraints) and inserting new ones. Rows existing in the current table
    but missing in the CSV won't be deleted.

    If `-r` is passed, the table will instead be deleted and recreated from the CSV file if it exists.
    """
    import csv

    from splitgraph.ingestion.csv import csv_adapter
    from splitgraph.ingestion.inference import infer_sg_schema

    if not primary_key:
        click.echo(
            "Warning: primary key is not specified, using the whole row as primary key."
            "This is probably not something that you want."
        )
    stream = ResettableStream(file)
    reader = csv.reader(io.TextIOWrapper(io.BufferedReader(stream)), delimiter=separator or ",")

    # Grab the first few rows from the CSV and give them to TableSchema.
    sample = list(islice(reader, 100))

    if no_header:
        # Patch in a dummy header
        sample = [[str(i) for i in range(len(sample))]] + sample

    type_overrides = dict(override_type or [])
    sg_schema = infer_sg_schema(
        sample,
        override_types=type_overrides,
        primary_keys=primary_key,
        ignore_empty_strings=ignore_empty_strings,
    )
    logging.debug("Using Splitgraph schema: %r", sg_schema)

    # Reset the stream and pass it to COPY FROM STDIN
    stream.reset()
    csv_adapter.to_table(
        io.TextIOWrapper(io.BufferedReader(stream)),
        repository,
        table,
        if_exists="replace" if replace else "patch",
        schema_check=not skip_schema_check,
        no_header=no_header,
        delimiter=separator,
        encoding=encoding,
        schema_spec=sg_schema,
    )


csv_group.add_command(csv_export)
csv_group.add_command(csv_import)



"""Public API for interacting with the Splitgraph registry"""
import base64
import contextlib
import json
import logging
import os
import time
import warnings
from functools import wraps
from json import JSONDecodeError
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import requests
from pydantic import BaseModel
from requests import HTTPError
from requests.adapters import HTTPAdapter
from requests.models import Response
from requests.packages.urllib3.util import Retry

from splitgraph.__version__ import __version__
from splitgraph.cloud.models import (
    AddExternalCredentialRequest,
    AddExternalRepositoriesRequest,
    AddExternalRepositoriesResponse,
    AddExternalRepositoryRequest,
    ExportJobStatus,
    ExternalResponse,
    IngestionJobStatus,
    IntrospectionMode,
    ListExternalCredentialsResponse,
    MetadataResponse,
    Plugin,
    RepositoryIngestionJobStatusResponse,
    UpdateExternalCredentialRequest,
    UpdateExternalCredentialResponse,
    make_repositories,
)
from splitgraph.cloud.project.models import External, Metadata, Repository
from splitgraph.cloud.queries import (
    BULK_UPDATE_REPO_SOURCES,
    BULK_UPSERT_REPO_PROFILES,
    BULK_UPSERT_REPO_TOPICS,
    CSV_URL,
    EXPORT_JOB_STATUS,
    FIND_REPO,
    GET_PLUGIN,
    GET_PLUGINS,
    GET_REPO_METADATA,
    GET_REPO_SOURCE,
    INGESTION_JOB_STATUS,
    JOB_LOGS,
    PROFILE_UPSERT,
    REPO_CONDITIONS,
    REPO_PARAMS,
    START_EXPORT,
    START_LOAD,
)
from splitgraph.config import CONFIG, create_config_dict, get_singleton
from splitgraph.config.config import (
    get_all_in_subsection,
    get_from_subsection,
    set_in_subsection,
)
from splitgraph.config.export import overwrite_config
from splitgraph.config.management import patch_and_save_config
from splitgraph.exceptions import (
    AuthAPIError,
    DataSourceError,
    GQLAPIError,
    GQLRepoDoesntExistError,
    GQLUnauthenticatedError,
    GQLUnauthorizedError,
    JSONSchemaValidationError,
)


def get_headers():
    return {"User-Agent": "sgr %s" % __version__}


DEFAULT_REMOTES = {
    "data.splitgraph.com": {
        "SG_IS_REGISTRY": "true",
        "SG_ENGINE_HOST": "data.splitgraph.com",
        "SG_ENGINE_PORT": "5432",
        "SG_ENGINE_DB_NAME": "sgregistry",
        "SG_AUTH_API": "https://api.splitgraph.com/auth",
        "SG_QUERY_API": "https://data.splitgraph.com",
        "SG_GQL_API": "https://api.splitgraph.com/gql/cloud/unified/graphql",
    }
}


def get_remote_param(remote: str, key: str) -> str:
    return str(
        get_all_in_subsection(CONFIG, "remotes", remote).get(key) or DEFAULT_REMOTES[remote][key]
    )


def expect_result(
    result: List[str], ignore_status_codes: Optional[List[int]] = None
) -> Callable[[Callable[..., Response]], Callable[..., Union[str, Tuple[str]]]]:
    """
    A decorator that can be wrapped around a function returning a requests.Response with a JSON body.
    If the request has failed, it will extract the "error" from the JSON response and raise an AuthAPIError.

    :param result: Items to extract. Will raise an AuthAPIError if not all items were fetched.
    :param ignore_status_codes: If one of these status codes is returned (e.g. 404),
        it gets ignored and a None is returned instead.
    :return: Tuple of items enumerated in the `result` list. If there's only one item, it will
        return just that item.
    """

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            try:
                response = func(*args, **kwargs)
            except Exception as e:
                raise AuthAPIError from e

            if ignore_status_codes and response.status_code in ignore_status_codes:
                return None

            try:
                response.raise_for_status()
            except HTTPError as e:
                error = response.text
                raise AuthAPIError(error) from e

            if not result:
                return None
            try:
                json = response.json()
            except JSONDecodeError:
                raise AuthAPIError("Invalid response from service: %s" % response.text)
            missing = [f for f in result if f not in json]
            if missing:
                raise AuthAPIError("Missing entries %s in the response!" % (tuple(missing),))
            if len(result) == 1:
                return json[result[0]]
            return tuple(json[f] for f in result)

        return wrapped

    return decorator


def _handle_gql_errors(response):
    logging.debug("GQL API status: %d, response: %s", response.status_code, response.text)
    if response.status_code not in (200, 400):
        response.raise_for_status()
    response_j = response.json()
    if "errors" in response_j:
        message = response_j["errors"][0]["message"]
        if "new row violates row-level security policy for table" in message:
            raise GQLUnauthorizedError("You do not have write access to this repository!")
        elif "Invalid token" in message:
            raise GQLUnauthenticatedError("Your access token doesn't exist or has expired!")
        elif "violates foreign key constraint" in message:
            raise GQLRepoDoesntExistError("Unknown repository!")
        else:
            raise GQLAPIError(message)
    # Catch other HTTP-level errors
    response.raise_for_status()


def handle_gql_errors(func: Callable[..., Response]) -> Callable[..., Response]:
    """
    A decorator that handles responses from the GQL API, transforming errors into exceptions.
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        response = func(*args, **kwargs)

        _handle_gql_errors(response)

        return response

    return wrapped


def get_token_claim(jwt, claim):
    """Extract a claim from a JWT token without validating it."""
    # Directly decode the base64 claims part without pulling in any JWT libraries
    # (since we're not validating any signatures).

    claims = jwt.split(".")[1]
    # Pad the JWT claims because urlsafe_b64decode doesn't like us
    claims += "=" * (-len(claims) % 4)
    exp = json.loads(base64.urlsafe_b64decode(claims).decode("utf-8"))[claim]
    return exp


T = TypeVar("T", bound=BaseModel)


class RESTAPIClient:
    """
    Client for various Splitgraph Registry REST APIs: auth token generation, external repo setup...
    """

    def __init__(self, remote: str) -> None:
        """
        :param remote: Name of the remote engine that this auth client communicates with,
            as specified in the config.
        """
        self.remote = remote
        self.endpoint = get_remote_param(remote, "SG_AUTH_API")
        self.externals_endpoint = get_remote_param(remote, "SG_QUERY_API") + "/api/external"

        # Allow overriding the CA bundle for test purposes (requests doesn't use the system
        # cert store)
        self.verify: Union[bool, str] = True
        with contextlib.suppress(KeyError):
            self.verify = get_remote_param(remote, "SG_AUTH_API_CA_PATH")

        # How soon before the token expiry to refresh the token, in seconds.
        self.access_token_expiry_tolerance = 30

    @expect_result(["tos"])
    def tos(self) -> Response:
        """
        Get a Terms of Service message from the registry (if accepting ToS is required)
        :return: Link to the Terms of Service or None
        """
        return requests.get(self.endpoint + "/tos", verify=self.verify, headers=get_headers())

    @expect_result(["user_id", "access_token", "refresh_token"])
    def register(self, username: str, password: str, email: str, accept_tos: bool) -> Response:
        """
        Register a new Splitgraph user.

        :param username: Username
        :param password: Password
        :param email: Email
        :param accept_tos: Accept the Terms of Service if they exist
        """
        body = {
            "username": username,
            "password": password,
            "email": email,
            "accept_tos": accept_tos,
        }

        headers = get_headers()
        with contextlib.suppress(AuthAPIError):
            # We can optionally pass an access token for logged-in admin users to make new users
            # on the registry if new signups are disabled, but it will be missing in most cases
            # (since the user is registering anew)
            headers["Authorization"] = "Bearer " + self.access_token
        return requests.post(
            self.endpoint + "/register_user", json=body, verify=self.verify, headers=headers
        )

    @expect_result(["access_token", "refresh_token"])
    def get_refresh_token(self, username: str, password: str) -> Response:
        """
        Get a long-lived refresh token and a short-lived access token from the API.

        :param username: Username
        :param password: Password
        :return: Tuple of (access_token, refresh_token).
        """
        body = {"username": username, "password": password}
        return requests.post(
            self.endpoint + "/refresh_token", json=body, verify=self.verify, headers=get_headers()
        )

    @expect_result(["key", "secret"])
    def create_machine_credentials(self, access_token: str, password: str) -> Response:
        """
        Generate a key and secret that can be used to log into the Splitgraph registry
        via a normal Postgres connection. The secret must be stored in the user's local
        configuration file (it's not stored on Splitgraph servers).

        :param access_token: Access token
        :param password: Password
        :return: Tuple of (key, secret).
        """
        body = {"password": password}
        return requests.post(
            self.endpoint + "/create_machine_credentials",
            json=body,
            headers={**get_headers(), **{"Authorization": "Bearer " + access_token}},
            verify=self.verify,
        )

    @expect_result(["access_token"])
    def get_access_token(self, refresh_token: str) -> Response:
        """
        Get a new access token from a refresh token.

        :param refresh_token: Refresh token
        :return: New access token.
        """

        body = {"refresh_token": refresh_token}
        return requests.post(
            self.endpoint + "/access_token", json=body, verify=self.verify, headers=get_headers()
        )

    @expect_result(["access_token"])
    def get_access_token_from_api(self, api_key: str, api_secret: str) -> Response:
        """
        Get a new access token from API keys

        :param api_key: API key
        :param api_secret: API secret
        :return: New access token.
        """

        body = {"api_key": api_key, "api_secret": api_secret}
        return requests.post(
            self.endpoint + "/access_token", json=body, verify=self.verify, headers=get_headers()
        )

    @property
    def access_token(self) -> str:
        """
        Will return an up-to-date access token by either getting it from
        the configuration file or contacting the auth service for a new one.
        Will write the new access token into the configuration file.

        :return: Access token.
        """

        config = create_config_dict()

        with contextlib.suppress(KeyError):
            current_access_token = get_from_subsection(
                config, "remotes", self.remote, "SG_CLOUD_ACCESS_TOKEN"
            )
            exp = get_token_claim(current_access_token, "exp")
            now = time.time()
            if now < exp - self.access_token_expiry_tolerance:
                return current_access_token

        # Token expired or non-existent, get a new one.
        try:
            api_key = get_from_subsection(config, "remotes", self.remote, "SG_ENGINE_USER")
            api_secret = get_from_subsection(config, "remotes", self.remote, "SG_ENGINE_PWD")
            new_access_token = cast(str, self.get_access_token_from_api(api_key, api_secret))
        except KeyError as e:
            try:
                refresh_token = get_from_subsection(
                    config, "remotes", self.remote, "SG_CLOUD_REFRESH_TOKEN"
                )
                new_access_token = cast(str, self.get_access_token(refresh_token))
            except KeyError:
                raise AuthAPIError(
                    (
                        "No refresh token or API keys found in the config for remote %s! "
                        % self.remote
                    )
                    + "Log into the registry using sgr cloud login."
                ) from e

        set_in_subsection(config, "remotes", self.remote, "SG_CLOUD_ACCESS_TOKEN", new_access_token)
        overwrite_config(config, get_singleton(config, "SG_CONFIG_FILE"))
        return new_access_token

    @property
    def maybe_access_token(self) -> Optional[str]:
        """
        Like access_token but returns None if the user isn't logged in.
        """
        try:
            return self.access_token
        except AuthAPIError:
            return None

    def get_latest_version(self) -> Optional[str]:
        # Do a version check to see if updates are available. If the user is logged
        # into the registry, also send the user ID for metrics.
        # The user can opt out by setting "SG_UPDATE_FREQUENCY" to 0 or opt out of
        # sending user ID by setting SG_UPDATE_ANONYMOUS to true.

        config = create_config_dict()
        frequency = int(get_singleton(config, "SG_UPDATE_FREQUENCY"))

        if frequency == 0:
            return None

        last_check = int(get_singleton(config, "SG_UPDATE_LAST"))
        now = int(time.time())

        if last_check + frequency > now:
            return None

        headers = get_headers()
        if get_singleton(config, "SG_UPDATE_ANONYMOUS").lower() == "false":
            with contextlib.suppress(AuthAPIError):
                headers.update({"Authorization": "Bearer " + self.access_token})

        try:
            logging.debug("Running update check")
            response = requests.post(
                self.endpoint + "/update_check",
                verify=self.verify,
                headers=headers,
            )
            response.raise_for_status()
            latest_version = str(response.json()["latest_version"])
        except requests.RequestException as e:
            logging.debug("Error running the update check", exc_info=e)
            return None
        except KeyError:
            logging.debug("Malformed response from the update service")
            return None

        try:
            patch_and_save_config(config, {"SG_UPDATE_LAST": str(now)})
        except Exception as e:
            logging.debug("Error patching the config", exc_info=e)
            return latest_version

        return latest_version

    def _perform_request(
        self,
        route,
        access_token,
        request: Optional[BaseModel] = None,
        response_class: Optional[Type[T]] = None,
        endpoint=None,
        method: str = "post",
        jsonschema_endpoint: bool = False,
    ) -> Optional[T]:
        endpoint = endpoint or self.endpoint
        response = requests.request(
            method,
            endpoint + route,
            headers={**get_headers(), **{"Authorization": "Bearer " + access_token}},
            verify=self.verify,
            data=request.json(by_alias=True, exclude_unset=True) if request else None,
        )
        logging.debug(response.text)

        if response.status_code in (405, 400) and jsonschema_endpoint:
            response_j = response.json()
            if "errors" in response_j:
                message = str(response_j["errors"])
            elif "error" in response_j:
                message = str(response_j["error"])
            else:
                message = response_j.text
            raise JSONSchemaValidationError(message)

        response.raise_for_status()

        if response_class:
            return response_class.parse_obj(response.json())
        return None

    def list_external_credentials(self) -> ListExternalCredentialsResponse:
        response = self._perform_request(
            "/list_external_credentials",
            self.access_token,
            None,
            ListExternalCredentialsResponse,
            method="get",
        )
        assert response
        return response

    def ensure_external_credential(
        self, credential_data: Dict[str, Any], credential_name: str, plugin_name: str
    ) -> str:
        """Store a credential for accessing an external data source
        in Splitgraph Cloud and return its ID"""
        access_token = self.access_token

        credentials = self.list_external_credentials()

        try:
            credential: Union[
                Optional["ListExternalCredentialsResponse.ExternalCredential"],
                Optional[UpdateExternalCredentialResponse],
            ]
            for credential in credentials.credentials:
                if (
                    credential.plugin_name == plugin_name
                    and credential.credential_name == credential_name
                ):
                    self._perform_request(
                        "/update_external_credential",
                        access_token,
                        UpdateExternalCredentialRequest(
                            credential_id=credential.credential_id,
                            credential_name=credential_name,
                            credential_data=credential_data,
                            plugin_name=plugin_name,
                        ),
                        jsonschema_endpoint=True,
                    )
                    return credential.credential_id

            # Credential doesn't exist: create it.
            credential = self._perform_request(
                "/add_external_credential",
                access_token,
                AddExternalCredentialRequest(
                    credential_name=credential_name,
                    credential_data=credential_data,
                    plugin_name=plugin_name,
                ),
                UpdateExternalCredentialResponse,
                jsonschema_endpoint=True,
            )

            assert credential
            return credential.credential_id
        except JSONSchemaValidationError:
            if logging.getLogger().getEffectiveLevel() >= logging.INFO:
                logging.error(
                    "JSONSchema error validating the credentials. Not displaying the contents of "
                    "the error, as it may contain the textual values of the credentials. "
                    "Run sgr with --verbosity DEBUG to see the full error message."
                )
                raise JSONSchemaValidationError(message="[MASKED]")
            raise

    def bulk_upsert_external(
        self,
        repositories: List[AddExternalRepositoryRequest],
        introspection_mode: IntrospectionMode = IntrospectionMode.EMPTY,
        raise_errors: bool = False,
    ) -> None:
        request = AddExternalRepositoriesRequest(
            repositories=repositories, introspection_mode=introspection_mode
        )
        response = self._perform_request(
            "/bulk-add",
            self.access_token,
            request,
            endpoint=self.externals_endpoint,
            jsonschema_endpoint=True,
            response_class=AddExternalRepositoriesResponse,
        )
        assert response
        if response.errors:
            for repo_errors in response.errors:
                for error in repo_errors.errors:
                    logging.warning(
                        "Error adding table %s/%s/%s: %s (%s)",
                        repo_errors.namespace,
                        repo_errors.repository,
                        error.table_name,
                        error.error,
                        error.error_text,
                    )
            if raise_errors:
                raise DataSourceError("Error introspecting some tables!")


def AuthAPIClient(*args, **kwargs):
    warnings.warn("AuthAPIClient is deprecated; use RESTAPIClient", DeprecationWarning, 2)
    return RESTAPIClient(*args, **kwargs)


class GQLAPIClient:
    """Wrapper class for select Splitgraph Registry GQL operations that can be
    called from the CLI"""

    def __init__(
        self,
        remote: Optional[str],
        endpoint: Optional[str] = None,
        access_token: Optional[str] = None,
    ):
        if not remote and not endpoint:
            raise ValueError(
                "GQLAPIClient must be initialized with either a remote or an endpoint!"
            )

        if remote:
            self.endpoint = get_remote_param(remote, "SG_GQL_API")
            if "/cloud/graphql" in self.endpoint:
                self.endpoint = self.endpoint.replace("/cloud/graphql", "/cloud/unified/graphql")
                logging.warning(
                    "The unified Splitgraph API is now at /gql/cloud/unified/graphql. "
                    "Using %s automatically. Replace SG_GQL_API to make this message go away. ",
                    self.endpoint,
                )
            self._auth_client: Optional[RESTAPIClient] = RESTAPIClient(remote)
            self._access_token: Optional[str] = access_token
        elif endpoint:
            self.endpoint = endpoint
            self._auth_client = None
            self._access_token = access_token

    @property
    def access_token(self) -> Optional[str]:
        if self._auth_client:
            return self._auth_client.access_token
        else:
            return self._access_token

    @property
    def maybe_access_token(self) -> Optional[str]:
        if self._auth_client:
            return self._auth_client.maybe_access_token
        else:
            return self._access_token

    def _gql(
        self, query: Dict, endpoint=None, handle_errors=False, anonymous_ok=False
    ) -> requests.Response:
        endpoint = endpoint or self.endpoint
        access_token = self.access_token if not anonymous_ok else self.maybe_access_token
        headers = get_headers()
        if access_token:
            headers.update({"Authorization": "Bearer " + access_token})

        # Add a retry strategy (difference from standard allowed_methods is that we also retry
        # on POST, since that's how we communicate with GQL endpoints).
        adapter = HTTPAdapter(
            max_retries=Retry(
                total=3,
                backoff_factor=1,
                status_forcelist=[429, 500, 502, 503, 504],
                allowed_methods=["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE", "POST"],
            )
        )
        session = requests.session()
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        result = session.post(
            endpoint, headers=headers, json=query, verify=os.environ.get("SSL_CERT_FILE", True)
        )
        if handle_errors:
            _handle_gql_errors(result)
        return result

    @staticmethod
    def _validate_metadata(namespace: str, repository: str, metadata: Metadata):
        # Pre-flight validation
        if metadata.description and len(metadata.description) > 160:
            raise ValueError("The description should be 160 characters or shorter!")

        variables: Dict[str, Any] = {"namespace": namespace, "repository": repository}

        variables.update(metadata.dict(by_alias=True, exclude_unset=True))

        variables = {k: v for k, v in variables.items() if v is not None}
        if "extra_metadata" in variables:
            extra_metadata = variables.pop("extra_metadata")

            # This is a bit of a hack. The actual metadata field in the repository is a JSON with
            # three toplevel fields:
            #   * created_at
            #   * updated_at
            #   * upstream_metadata
            #
            # The former two are used to populate schema.org specifications and override the
            # Updated At field on the repository page. The latter is used to render the
            # semi-structured box at the bottom with double nesting and arbitrary metadata.
            # We don't necessarily want to keep it that way (it was designed to support cataloguing
            # upstream open datasets), but want to let users set these anyway. Here, we
            # pluck out the two special "created_at"/"updated_at" fields into the toplevel
            # and put the rest of the metadata dict into "upstream_metadata" to get it rendering.
            metadata_doc: Dict[str, Any] = {}
            for toplevel_key in ["created_at", "updated_at"]:
                if toplevel_key in extra_metadata:
                    metadata_doc[toplevel_key] = str(extra_metadata.pop(toplevel_key))
            metadata_doc["upstream_metadata"] = extra_metadata
            variables["metadata"] = metadata_doc

        if "readme" in variables and isinstance(variables["readme"], dict):
            variables["readme"] = variables["readme"]["text"]

        return variables

    @staticmethod
    def _prepare_upsert_metadata_gql(namespace: str, repository: str, metadata: Metadata):
        variables = GQLAPIClient._validate_metadata(namespace, repository, metadata)

        query = {
            "operationName": "UpsertRepoProfile",
            "variables": variables,
            "query": PROFILE_UPSERT,
        }

        logging.debug("Prepared GraphQL query: %s", json.dumps(query))

        return query

    @handle_gql_errors
    def upsert_metadata(self, namespace: str, repository: str, metadata: Metadata):
        """
        Update metadata for a single repository.
        """
        response = self._gql(self._prepare_upsert_metadata_gql(namespace, repository, metadata))
        return response

    def bulk_upsert_metadata(
        self, namespace_list: List[str], repository_list: List[str], metadata_list: List[Metadata]
    ):
        repo_profiles: Dict[str, List[Any]] = {
            "namespaces": namespace_list,
            "repositories": repository_list,
            "descriptions": [],
            "readmes": [],
            "licenses": [],
            "metadata": [],
        }
        repo_sources: Dict[str, List[Any]] = {"namespaces": [], "repositories": [], "sources": []}
        repo_topics: Dict[str, List[str]] = {"namespaces": [], "repositories": [], "topics": []}

        # populate mutation payloads
        for ind, metadata in enumerate(metadata_list):
            validated_metadata = GQLAPIClient._validate_metadata(
                namespace_list[ind], repository_list[ind], metadata
            )

            repo_profiles["descriptions"].append(validated_metadata.get("description"))
            repo_profiles["readmes"].append(validated_metadata.get("readme"))
            repo_profiles["licenses"].append(validated_metadata.get("license"))
            repo_profiles["metadata"].append(validated_metadata.get("metadata"))

            # flatten sources, which will be aggregated on the server side
            if len(validated_metadata.get("sources", [])) > 0:
                for source in validated_metadata["sources"]:
                    repo_sources["namespaces"].append(namespace_list[ind])
                    repo_sources["repositories"].append(repository_list[ind])
                    repo_sources["sources"].append(source)

            # flatten topics, which will be aggregated on the server side
            if len(validated_metadata.get("topics", [])) > 0:
                for topic in validated_metadata["topics"]:
                    repo_topics["namespaces"].append(namespace_list[ind])
                    repo_topics["repositories"].append(repository_list[ind])
                    repo_topics["topics"].append(topic)

        self._bulk_upsert_repo_profiles(repo_profiles)
        self._bulk_upsert_repo_sources(repo_sources)
        self._bulk_upsert_repo_topics(repo_topics)

    @handle_gql_errors
    def _bulk_upsert_repo_profiles(self, repo_profiles: Dict[str, List[Any]]):
        repo_profiles_query = {
            "operationName": "BulkUpsertRepoProfilesMutation",
            "variables": repo_profiles,
            "query": BULK_UPSERT_REPO_PROFILES,
        }
        response = self._gql(repo_profiles_query)
        return response

    @handle_gql_errors
    def _bulk_upsert_repo_sources(self, repo_sources: Dict[str, List[Any]]):
        repo_sources_query = {
            "operationName": "BulkUpdateRepoSourcesMutation",
            "variables": repo_sources,
            "query": BULK_UPDATE_REPO_SOURCES,
        }
        response = self._gql(repo_sources_query)
        return response

    @handle_gql_errors
    def _bulk_upsert_repo_topics(self, repo_topics: Dict[str, List[str]]):
        repo_topics_query = {
            "operationName": "BulkUpsertRepoTopicsMutation",
            "variables": repo_topics,
            "query": BULK_UPSERT_REPO_TOPICS,
        }
        response = self._gql(repo_topics_query)
        return response

    def upsert_readme(self, namespace: str, repository: str, readme: str):
        return self.upsert_metadata(namespace, repository, Metadata(readme=readme))

    @handle_gql_errors
    def upsert_description(self, namespace: str, repository: str, description: str):
        return self.upsert_metadata(namespace, repository, Metadata(description=description))

    @handle_gql_errors
    def upsert_topics(self, namespace: str, repository: str, topics: List[str]):
        return self.upsert_metadata(namespace, repository, Metadata(topics=topics))

    def find_repository(
        self, query: str, limit: int = 10
    ) -> Tuple[int, List[Tuple[str, str, str]]]:
        response = self._gql(
            {
                "operationName": "FindRepositories",
                "variables": {"query": query, "limit": limit},
                "query": FIND_REPO,
            }
        )

        _handle_gql_errors(response)
        result = response.json()

        # Extract data from the response
        find_repository = result["data"]["findRepository"]
        total_count = find_repository["totalCount"]
        repos_previews = [
            (r["node"]["namespace"], r["node"]["repository"], r["node"]["highlight"])
            for r in find_repository["edges"]
        ]

        return total_count, repos_previews

    def get_metadata(self, namespace: str, repository: str) -> Optional[MetadataResponse]:
        response = self._gql(
            {
                "query": GET_REPO_METADATA % (REPO_PARAMS, REPO_CONDITIONS),
                "operationName": "GetRepositoryMetadata",
                "variables": {"namespace": namespace, "repository": repository},
            },
            handle_errors=True,
        )

        parsed_responses = MetadataResponse.from_response(response.json())
        if parsed_responses:
            assert len(parsed_responses) == 1
            return parsed_responses[0]
        return None

    def get_latest_ingestion_job_status(
        self, namespace: str, repository: str
    ) -> Optional[IngestionJobStatus]:
        response = self._gql(
            {
                "query": INGESTION_JOB_STATUS,
                "operationName": "RepositoryIngestionJobStatus",
                "variables": {"namespace": namespace, "repository": repository},
            },
            handle_errors=True,
        )
        parsed_response = RepositoryIngestionJobStatusResponse.from_response(response.json())
        nodes = parsed_response.repositoryIngestionJobStatus.nodes
        if not nodes:
            return None
        else:
            assert len(nodes) == 1
            node = nodes[0]
            return IngestionJobStatus(
                task_id=node.taskId,
                started=node.started,
                finished=node.finished,
                is_manual=node.isManual,
                status=node.status,
            )

    def get_export_job_status(self, task_id: str) -> Optional[ExportJobStatus]:
        response = self._gql(
            {
                "query": EXPORT_JOB_STATUS,
                "operationName": "ExportJobStatus",
                "variables": {"taskId": task_id},
            },
            handle_errors=True,
        )

        data = response.json()["data"]["exportJobStatus"]
        if not data:
            return None
        return ExportJobStatus(
            task_id=data["taskId"],
            started=data["started"],
            finished=data["finished"],
            status=data["status"],
            user_id=data["userId"],
            export_format=data["exportFormat"],
            output=data["output"],
        )

    def start_export(self, query: str) -> str:
        query = query.strip()
        if query.endswith(";"):
            logging.warning("The query ends with ';', automatically removing")
            query = query[:-1]

        response = self._gql(
            {
                "query": START_EXPORT,
                "operationName": "StartExport",
                "variables": {"query": query},
            },
            handle_errors=True,
        )
        return str(response.json()["data"]["exportQuery"]["id"])

    def get_ingestion_job_logs(self, namespace: str, repository: str, task_id: str) -> str:
        response = self._gql(
            {
                "query": JOB_LOGS,
                "operationName": "JobLogs",
                "variables": {"namespace": namespace, "repository": repository, "taskId": task_id},
            },
            handle_errors=True,
        )

        url = response.json()["data"]["jobLogs"]["url"]

        if not url:
            raise ValueError(
                "Task ID %s not found for repository %s/%s!" % (task_id, namespace, repository)
            )

        response = requests.get(url)
        response.raise_for_status()
        return response.text

    def get_csv_upload_download_urls(self) -> Tuple[str, str]:
        response = self._gql(
            {
                "query": CSV_URL,
                "operationName": "CSVURLs",
            },
            handle_errors=True,
        )
        urls = response.json()["data"]["csvUploadDownloadUrls"]
        return urls["upload"], urls["download"]

    def _run_start_load_gql_with(self, variables: Dict[str, Any]) -> str:
        response = self._gql(
            {
                "query": START_LOAD,
                "operationName": "StartExternalRepositoryLoad",
                "variables": variables,
            },
            handle_errors=True,
        )
        return str(response.json()["data"]["startExternalRepositoryLoad"]["taskId"])

    def start_csv_load(
        self, namespace: str, repository: str, download_urls: List[str], table_names: List[str]
    ) -> str:
        variables = {
            "namespace": namespace,
            "repository": repository,
            "pluginName": "csv",
            "params": json.dumps({"connection": {"connection_type": "http", "url": ""}}),
            "tableParams": [
                {
                    "name": n,
                    "options": json.dumps({"url": u}),
                    "schema": [],
                }
                for n, u in zip(table_names, download_urls)
            ],
        }

        return self._run_start_load_gql_with(variables)

    @staticmethod
    def _make_plugin(plugin_dict: Dict[str, Any]) -> Plugin:
        return Plugin(
            plugin_name=plugin_dict["pluginName"],
            credentials_schema=plugin_dict["credentialsSchema"],
            params_schema=plugin_dict["paramsSchema"],
            table_params_schema=plugin_dict["tableParamsSchema"],
            name=plugin_dict["name"],
            description=plugin_dict["description"],
            supports_load=plugin_dict["supportsLoad"],
            supports_sync=plugin_dict["supportsSync"],
            supports_mount=plugin_dict["supportsMount"],
        )

    def get_all_plugins(self, seed: Optional[str] = None) -> List[Plugin]:
        response = self._gql(
            {"query": GET_PLUGINS, "operationName": "ExternalPlugins", "variables": {"seed": seed}},
            handle_errors=True,
            anonymous_ok=True,
        )
        return [self._make_plugin(d) for d in response.json()["data"]["externalPlugins"]]

    def get_plugin(self, plugin_name: str) -> Optional[Plugin]:
        response = self._gql(
            {
                "query": GET_PLUGIN,
                "operationName": "ExternalPlugin",
                "variables": {"pluginName": plugin_name},
            },
            handle_errors=True,
            anonymous_ok=True,
        )
        data = response.json()["data"]["externalPlugin"]
        if data is None:
            return None
        return self._make_plugin(data)

    def start_load_existing(self, namespace: str, repository: str, sync: bool = True) -> str:
        variables = {"namespace": namespace, "repository": repository, "sync": sync}

        return self._run_start_load_gql_with(variables)

    def start_load_params(
        self,
        namespace: str,
        repository: str,
        external: External,
        sync: bool = True,
        credential_data: Optional[Dict[str, Any]] = None,
        initial_private: bool = False,
    ) -> str:
        variables = {
            "namespace": namespace,
            "repository": repository,
            "pluginName": external.plugin,
            "params": json.dumps(external.params),
            "tableParams": [
                {
                    "name": tn,
                    "options": json.dumps(to.options),
                    "schema": [{"name": tc.name, "pgType": tc.pg_type} for tc in to.schema_],
                }
                for tn, to in external.tables.items()
            ],
            "sync": sync,
            "initialVisibility": ("PRIVATE" if initial_private else "PUBLIC"),
        }

        if credential_data:
            variables["credentialData"] = json.dumps(credential_data)
        elif external.credential_id:
            variables["credentialId"] = external.credential_id
        return self._run_start_load_gql_with(variables)

    def get_external_metadata(self, namespace: str, repository: str) -> Optional[ExternalResponse]:
        response = self._gql(
            {
                "query": GET_REPO_SOURCE % (REPO_PARAMS, REPO_CONDITIONS),
                "operationName": "GetRepositoryDataSource",
                "variables": {"namespace": namespace, "repository": repository},
            },
            handle_errors=True,
        )

        parsed_responses = ExternalResponse.from_response(response.json())
        if parsed_responses:
            assert len(parsed_responses) == 1
            return parsed_responses[0]
        return None

    def load_all_repositories(self, limit_to: List[str] = None) -> List[Repository]:
        from splitgraph.core.repository import Repository as SGRepository

        if limit_to:
            parsed_metadata = []
            parsed_external = []

            for repository in limit_to:
                # We currently can't filter on multiple fields in GQL, so we loop instead.
                repo_obj = SGRepository.from_schema(repository)
                metadata = self.get_metadata(repo_obj.namespace, repo_obj.repository)
                external = self.get_external_metadata(repo_obj.namespace, repo_obj.repository)

                if metadata:
                    parsed_metadata.append(metadata)
                if external:
                    parsed_external.append(external)
        else:
            metadata_r = self._gql(
                {
                    "query": GET_REPO_METADATA % ("", ""),
                    "operationName": "GetRepositoryMetadata",
                },
                handle_errors=True,
            )

            external_r = self._gql(
                {
                    "query": GET_REPO_SOURCE % ("", ""),
                    "operationName": "GetRepositoryDataSource",
                },
                handle_errors=True,
            )
            parsed_metadata = MetadataResponse.from_response(metadata_r.json())
            parsed_external = ExternalResponse.from_response(external_r.json())

        return make_repositories(parsed_metadata, parsed_external)



"""
Splitgraph command line client

Hooks into the API to allow management of Splitgraph repositories and images using ``sgr``.
"""

import logging
import traceback

import click
import click_log
from click_log import ColorFormatter

from splitgraph.__version__ import __version__
from splitgraph.commandline.cloud import cloud_c
from splitgraph.commandline.engine import engine_c
from splitgraph.commandline.example import example
from splitgraph.commandline.image_creation import (
    checkout_c,
    commit_c,
    import_c,
    reindex_c,
    tag_c,
)
from splitgraph.commandline.image_info import (
    diff_c,
    log_c,
    object_c,
    objects_c,
    show_c,
    sql_c,
    status_c,
    table_c,
)
from splitgraph.commandline.ingestion import csv_group
from splitgraph.commandline.misc import (
    cleanup_c,
    config_c,
    dump_c,
    eval_c,
    init_c,
    prune_c,
    rm_c,
    upgrade_c,
)
from splitgraph.commandline.mount import mount_c
from splitgraph.commandline.push_pull import clone_c, pull_c, push_c, upstream_c
from splitgraph.commandline.splitfile import (
    build_c,
    dependents_c,
    provenance_c,
    rebuild_c,
)
from splitgraph.exceptions import get_exception_name
from splitgraph.ingestion.singer.commandline import singer_group

logger = logging.getLogger()


class ClickHandler(logging.Handler):
    def emit(self, record):
        try:
            msg = self.format(record)
            click.echo(msg, err=True)
        except Exception:
            self.handleError(record)


click_log.basic_config(logger)
logger.handlers = [ClickHandler()]
logger.handlers[0].formatter = ColorFormatter()


def _patch_wrap_text():
    # Patch click's formatter to strip ```
    # etc which are used to format help text on the website.
    wrap_text = click.formatting.wrap_text

    def patched_wrap_text(*args, **kwargs):
        text = args[0]
        text = text.replace("```\n", "").replace("`", "")

        return wrap_text(text, *args[1:], **kwargs)

    click.formatting.wrap_text = patched_wrap_text


def _do_version_check():
    """Do a pre-flight version check -- by default we only do it once a day"""
    from packaging.version import Version

    from splitgraph.cloud import RESTAPIClient
    from splitgraph.config import CONFIG, get_singleton

    api_client = RESTAPIClient(get_singleton(CONFIG, "SG_UPDATE_REMOTE"))
    latest = api_client.get_latest_version()

    if not latest:
        return

    if Version(latest) > Version(__version__):
        click.echo(
            "You are using sgr version %s, however version %s is available." % (__version__, latest)
        )
        click.echo("Consider upgrading by running sgr upgrade or pip install -U splitgraph.")
        click.echo("Disable this message by setting SG_UPDATE_FREQUENCY=0 in your .sgconfig.")


class WithExceptionHandler(click.Group):
    def get_command(self, ctx, cmd_name):
        # Patch only if we're invoked somehow, not if we're imported.
        _patch_wrap_text()
        return super().get_command(ctx, cmd_name)

    def invoke(self, ctx):
        import psycopg2.extensions
        import psycopg2.extras

        from splitgraph.engine import get_engine

        # Allow users to send SIGINT to quickly terminate sgr (instead of waiting for a PG
        # statement to finish)
        psycopg2.extensions.set_wait_callback(psycopg2.extras.wait_select)

        engine = get_engine()
        try:
            result = super(click.Group, self).invoke(ctx)
            engine.commit()
            return result
        except Exception as exc:
            engine.rollback()
            if isinstance(
                exc,
                (click.exceptions.ClickException, click.exceptions.Abort, click.exceptions.Exit),
            ):
                raise
            # Can't seem to be able to get the click_log verbosity option
            # value so have to get it indirectly. Basically, if we're in
            # DEBUG mode, output the whole stacktrace.
            if logger.getEffectiveLevel() == logging.DEBUG:
                logger.error(traceback.format_exc())
            else:
                logger.error("%s: %s" % (get_exception_name(exc), exc))
            ctx.exit(code=2)
        finally:
            _do_version_check()
            engine.close()


@click.group(cls=WithExceptionHandler)
@click_log.simple_verbosity_option(logger)
@click.version_option(prog_name="sgr", version=__version__)
def cli():
    """Splitgraph command line client: manage and build Postgres schema images."""


# Note on the docstring format:
# * Click uses the first sentence as a short help text as a command group
# * All docstrings are in the imperative mood
#   (e.g. "Commit a Splitgraph schema" instead of "Commits a Splitgraph schema".)

# Note on performance:
# To avoid Python's famously slow imports which tank the performance of something like sgr --help,
# none of the CLI modules import the rest of splitgraph at toplevel apart from lightweight helpers:
# splitgraph.config and splitgraph.exceptions.
# All other imports are recommended to be pushed into the actual command definition.

# Possible extra commands:
# * sgr squash namespace/repo:image: takes an image, turns all of its objects into snapshots and creates
#   a new image (useful for publishing?)
# * sgr reset namespace/repo:image: similar to a git "soft reset": moves the HEAD pointer to the target
#   and restages all changes between the new target and the old HEAD.
#
#   It might be possible to implement this by not suspending the audit trigger on the target tables whilst
#   rerunning the checkout (applying all objects), which should put the staging area in the same state
#   and recalculate the correct pending changes for the audit table.

# Image management/creation
cli.add_command(checkout_c)
cli.add_command(commit_c)
cli.add_command(tag_c)
cli.add_command(import_c)
cli.add_command(reindex_c)

# Information
cli.add_command(log_c)
cli.add_command(diff_c)
cli.add_command(object_c)
cli.add_command(objects_c)
cli.add_command(show_c)
cli.add_command(table_c)
cli.add_command(sql_c)
cli.add_command(status_c)

# Engine management
cli.add_command(engine_c)

# Miscellaneous
cli.add_command(mount_c)
cli.add_command(rm_c)
cli.add_command(init_c)
cli.add_command(cleanup_c)
cli.add_command(prune_c)
cli.add_command(config_c)
cli.add_command(dump_c)
cli.add_command(eval_c)
cli.add_command(upgrade_c)

# Push/pull/sharing
cli.add_command(clone_c)
cli.add_command(pull_c)
cli.add_command(push_c)
cli.add_command(upstream_c)

# Splitfile execution
cli.add_command(build_c)
cli.add_command(dependents_c)
cli.add_command(provenance_c)
cli.add_command(rebuild_c)

# Examples
cli.add_command(example)

# CSV
cli.add_command(csv_group)

# Cloud
cli.add_command(cloud_c)

# Singer
cli.add_command(singer_group)



"""
Hooks for registering handlers to upload/download objects from external locations into Splitgraph's cache.
"""
from importlib import import_module
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Sequence, Tuple

from splitgraph.config import CONFIG
from splitgraph.config.config import get_from_section
from splitgraph.exceptions import ExternalHandlerError

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PsycopgEngine


class ExternalObjectHandler:
    """
    Framework for allowing to dump objects from the Splitgraph cache to an external location. This allows
    the objects to be stored somewhere other than the actual remote engine.

    External object handlers must extend this class and be registered in the Splitgraph config.

    For an example of how this can be used, see splitgraph.hooks.s3: it's a handler allowing objects to be
    uploaded to S3/S3-compatible host using the Minio API. It's registered in the config as follows::

        [external_handlers]
        S3=splitgraph.hooks.s3.S3ExternalObjectHandler

    The protocol and the URLs returned by this handler are stored in splitgraph_meta.external_objects
    and used to download the objects back into the Splitgraph cache when they are needed.
    """

    def __init__(self, params: Dict[Any, Any]) -> None:
        """
        Initialize the handler with a dictionary of parameters (e.g. remote access keys etc).
        The instantiation happens at the beginning of a pull/push.

        :param params: Extra params to be passed to the handler from the commandline invocation
            of push/pull.
        """
        self.params = params

    def upload_objects(
        self, objects: List[str], remote_engine: "PsycopgEngine"
    ) -> Sequence[Tuple[str, str]]:
        """Upload objects from the Splitgraph cache to an external location

        :param objects: List of object IDs to upload
        :param remote_engine: An instance of Engine class that the objects will be registered on
        :return: A list of successfully uploaded object IDs and URLs they can be found at.
        """

    def download_objects(
        self, objects: List[Tuple[str, str]], remote_engine: "PsycopgEngine"
    ) -> Sequence[str]:
        """Download objects from the external location into the Splitgraph cache.

        :param objects: List of tuples `(object_id, object_url)` that this handler had previosly
            uploaded the objects to.
        :param remote_engine: An instance of Engine class that the objects will be registered on
        :return: A list of object IDs that have been successfully downloaded.
        """


_EXTERNAL_OBJECT_HANDLERS: Dict[str, Callable[..., ExternalObjectHandler]] = {}


def get_external_object_handler(name: str, handler_params: Dict[Any, Any]) -> ExternalObjectHandler:
    """Load an external protocol handler by its name, initializing it with optional parameters."""
    try:
        handler_class = _EXTERNAL_OBJECT_HANDLERS[name]
        return handler_class(handler_params)
    except KeyError:
        try:
            handler_class_name = get_from_section(CONFIG, "external_handlers", name)
        except KeyError:
            raise ExternalHandlerError("Protocol %s is not supported!" % name)

        index = handler_class_name.rindex(".")
        try:
            handler_class = getattr(
                import_module(handler_class_name[:index]), handler_class_name[index + 1 :]
            )
            register_upload_download_handler(name, handler_class)
            return handler_class(handler_params)
        except AttributeError as e:
            raise ExternalHandlerError(
                "Error loading external object handler {0}".format(name)
            ) from e
        except ImportError as e:
            raise ExternalHandlerError(
                "Error loading external object handler {0}".format(name)
            ) from e


def register_upload_download_handler(
    name: str, handler_class: Callable[..., ExternalObjectHandler]
) -> None:
    """Register an external protocol handler. See the docstring for `get_upload_download_handler` for the required
    signatures of the handler functions."""
    global _EXTERNAL_OBJECT_HANDLERS
    if name in _EXTERNAL_OBJECT_HANDLERS:
        raise ExternalHandlerError(
            "Cannot register a protocol handler %s as it already exists!" % name
        )
    _EXTERNAL_OBJECT_HANDLERS[name] = handler_class



"""Image representation and provenance"""

import logging
from contextlib import contextmanager
from datetime import datetime
from random import getrandbits
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    cast,
)

from psycopg2.extras import Json
from psycopg2.sql import SQL, Identifier

from splitgraph.config import (
    CONFIG,
    FDW_CLASS,
    SPLITGRAPH_API_SCHEMA,
    SPLITGRAPH_META_SCHEMA,
    get_singleton,
)
from splitgraph.engine import ResultShape
from splitgraph.exceptions import SplitGraphError, TableNotFoundError

from .common import manage_audit, set_head, set_tag, unmount_schema
from .sql import POSTGRES_MAX_IDENTIFIER, prepare_splitfile_sql, select
from .table import Table
from .types import ProvenanceLine, TableColumn

if TYPE_CHECKING:
    from .repository import Repository

IMAGE_COLS = ["image_hash", "parent_id", "created", "comment", "provenance_data"]


class Image(NamedTuple):
    """
    Represents a Splitgraph image. Should't be created directly, use Image-loading methods in the
    :class:`splitgraph.core.repository.Repository` class instead.
    """

    image_hash: str
    parent_id: Optional[str]
    created: datetime
    comment: str
    provenance_data: List[ProvenanceLine]
    repository: "Repository"

    @property
    def engine(self):
        return self.repository.engine

    @property
    def object_engine(self):
        return self.repository.object_engine

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Image):
            return NotImplemented
        return self.image_hash == other.image_hash and self.repository == other.repository

    def get_parent_children(self) -> Tuple[Optional[str], List[str]]:
        """Gets the parent and a list of children of a given image."""
        parent = self.parent_id

        children = self.engine.run_sql(
            SQL("SELECT image_hash FROM {}.get_images(%s,%s) WHERE parent_id = %s").format(
                Identifier(SPLITGRAPH_API_SCHEMA)
            ),
            (self.repository.namespace, self.repository.repository, self.image_hash),
            return_shape=ResultShape.MANY_ONE,
        )
        return parent, children

    def get_tables(self) -> List[str]:
        """
        Gets the names of all tables inside of an image.
        """
        result = self.engine.run_sql(
            select(
                "get_tables", "table_name", table_args="(%s,%s,%s)", schema=SPLITGRAPH_API_SCHEMA
            ),
            (self.repository.namespace, self.repository.repository, self.image_hash),
            return_shape=ResultShape.MANY_ONE,
        )
        return result or []

    def get_table(self, table_name: str) -> Table:
        """
        Returns a Table object representing a version of a given table.
        Contains a list of objects that the table is linked to and the table's schema.

        :param table_name: Name of the table
        :return: Table object
        """
        result = self.engine.run_sql(
            select(
                "get_tables",
                "table_schema,object_ids",
                "table_name = %s",
                table_args="(%s,%s,%s)",
                schema=SPLITGRAPH_API_SCHEMA,
            ),
            (self.repository.namespace, self.repository.repository, self.image_hash, table_name),
            return_shape=ResultShape.ONE_MANY,
        )
        if not result:
            raise TableNotFoundError(
                "Image %s:%s does not have a table %s!"
                % (self.repository, self.image_hash, table_name)
            )
        table_schema, objects = result
        return Table(
            self.repository, self, table_name, [TableColumn(*t) for t in table_schema], objects
        )

    @manage_audit
    def checkout(self, force: bool = False, layered: bool = False) -> None:
        """
        Checks the image out, changing the current HEAD pointer. Raises an error
        if there are pending changes to its checkout.

        :param force: Discards all pending changes to the schema.
        :param layered: If True, uses layered querying to check out the image (doesn't materialize tables
            inside of it).
        """
        target_schema = self.repository.to_schema()
        if len(target_schema) > POSTGRES_MAX_IDENTIFIER:
            logging.warning(
                "The full repository name %s is longer than PostgreSQL's maximum "
                "identifier length of %d. PostgreSQL will truncate the schema name "
                "down to %d characters in all queries, which might cause clashes "
                "with other checked-out repositories.",
                target_schema,
                POSTGRES_MAX_IDENTIFIER,
                POSTGRES_MAX_IDENTIFIER,
            )

        if self.repository.has_pending_changes():
            if not force:
                raise SplitGraphError(
                    "{0} has pending changes! Pass force=True or do sgr checkout -f {0}:HEAD".format(
                        target_schema
                    )
                )
            logging.warning("%s has pending changes, discarding...", target_schema)
            self.object_engine.discard_pending_changes(target_schema)

        # Drop all current tables in staging
        self.object_engine.create_schema(target_schema)
        for table in self.object_engine.get_all_tables(target_schema):
            self.object_engine.delete_table(target_schema, table)

        if layered:
            self.lq_checkout()
        else:
            for table in self.get_tables():
                self.get_table(table).materialize(table)
        set_head(self.repository, self.image_hash)

    def lq_checkout(
        self,
        target_schema: Optional[str] = None,
        wrapper: Optional[str] = FDW_CLASS,
        only_tables: Optional[List[str]] = None,
    ) -> None:
        """
        Intended to be run on the sgr side. Initializes the FDW for all tables in a given image,
        allowing to query them directly without materializing the tables.
        """
        # assumes that we got to the point in the normal checkout where we're about to materialize the tables
        # (e.g. the schemata are cleared)
        # Use a per-schema "foreign server" for layered queries for now

        # Circular import
        from splitgraph.hooks.data_source.fdw import init_fdw

        target_schema = target_schema or self.repository.to_schema()
        server_id = "%s_lq_checkout_server" % target_schema
        engine = self.repository.engine
        object_engine = self.repository.object_engine

        init_fdw(
            object_engine,
            server_id=server_id,
            wrapper="multicorn",
            server_options={
                "wrapper": wrapper,
                "engine": engine.name,
                "object_engine": object_engine.name,
                "namespace": self.repository.namespace,
                "repository": self.repository.repository,
                "image_hash": self.image_hash,
            },
        )

        # It's easier to create the foreign tables from our side than to implement IMPORT FOREIGN SCHEMA by the FDW
        for table_name in self.get_tables():
            if only_tables and table_name not in only_tables:
                continue

            logging.debug(
                "Mounting %s:%s/%s into %s",
                self.repository.to_schema(),
                self.image_hash,
                table_name,
                target_schema,
            )
            self.get_table(table_name).materialize(table_name, target_schema, lq_server=server_id)

    @contextmanager
    def query_schema(
        self, wrapper: Optional[str] = FDW_CLASS, commit: bool = True
    ) -> Iterator[str]:
        """
        Creates a temporary schema with tables in this image mounted as foreign tables that can be accessed via
        read-only layered querying. On exit from the context manager, the schema is discarded.

        :return: The name of the schema the image is located in.
        """
        tmp_schema = str.format("o{:032x}", getrandbits(128))
        try:
            self.object_engine.create_schema(tmp_schema)
            self.lq_checkout(target_schema=tmp_schema, wrapper=wrapper)
            if commit:
                self.object_engine.commit()  # Make sure the new tables are seen by other connections

            # Inject extra query planner hints as session variables if specified.
            lq_tuning = get_singleton(CONFIG, "SG_LQ_TUNING")
            if lq_tuning:
                self.object_engine.run_sql(lq_tuning)
            yield tmp_schema
        finally:
            unmount_schema(self.object_engine, tmp_schema)

    def tag(self, tag: str) -> None:
        """
        Tags a given image. All tags are unique inside of a repository. If a tag already exists, it's removed
        from the previous image and given to the new image.

        :param tag: Tag to set. 'latest' and 'HEAD' are reserved tags.
        """
        set_tag(self.repository, self.image_hash, tag)

    def get_tags(self):
        """Lists all tags that this image has."""
        return [t for h, t in self.repository.get_all_hashes_tags() if h == self.image_hash]

    def delete_tag(self, tag: str) -> None:
        """
        Deletes a tag from an image.

        :param tag: Tag to delete.
        """

        # Does checks to make sure the tag actually exists, will raise otherwise
        self.repository.images.by_tag(tag)

        self.engine.run_sql(
            select("delete_tag", table_args="(%s,%s,%s)", schema=SPLITGRAPH_API_SCHEMA),
            (self.repository.namespace, self.repository.repository, tag),
            return_shape=None,
        )

    def get_log(self) -> List["Image"]:
        """Repeatedly gets the parent of a given image until it reaches the bottom."""
        all_images = {i.image_hash: i for i in self.repository.images()}
        result = [self]
        while result[-1].parent_id is not None:
            try:
                result.append(all_images[result[-1].parent_id])
            except KeyError:
                # If we don't have the parent's metadata, it's possible
                # that the parent hasn't been pulled -- ignore it and stop here.
                return result
        return result

    def get_size(self) -> int:
        """
        Get the physical size used by the image's objects (including those that might be
        shared with other images).

        This is calculated from the metadata, the on-disk footprint might be smaller if not all of image's
        objects have been downloaded.

        :return: Size of the image in bytes.
        """
        return cast(
            int,
            self.engine.run_sql(
                select("get_image_size", table_args="(%s,%s,%s)", schema=SPLITGRAPH_API_SCHEMA),
                (self.repository.namespace, self.repository.repository, self.image_hash),
                return_shape=ResultShape.ONE_ONE,
            )
            or 0,
        )

    def to_splitfile(
        self,
        ignore_irreproducible: bool = False,
        source_replacement: Optional[Dict["Repository", str]] = None,
    ) -> List[str]:
        """
        Recreate the Splitfile that can be used to reconstruct this image.

        :param ignore_irreproducible: If True, ignore commands from irreproducible Splitfile lines
            (like MOUNT or custom commands) and instead emit a comment (this results in an invalid Splitfile).
        :param source_replacement: A dictionary of repositories and image hashes/tags specifying how to replace the
            dependencies of this Splitfile (table imports and FROM commands).
        :return: A list of Splitfile commands that can be fed back into the executor.
        """

        return reconstruct_splitfile(
            self.provenance_data, ignore_irreproducible, source_replacement
        )

    def provenance(self, reverse=False, engine=None) -> List[Tuple["Repository", str]]:
        """
        Inspects the image's parent chain to come up with a set of repositories and their hashes
        that it was created from.

        If `reverse` is True, returns a list of images that were created _from_ this image. If
        this image is on a remote repository, `engine` can be passed in to override the engine
        used for the lookup of dependents.

        :return: List of (repository, image_hash)
        """
        from splitgraph.core.repository import Repository

        api_call = "get_image_dependents" if reverse else "get_image_dependencies"

        engine = engine or self.engine

        result = set()
        for namespace, repository, image_hash in engine.run_sql(
            select(api_call, table_args="(%s,%s,%s)", schema=SPLITGRAPH_API_SCHEMA),
            (self.repository.namespace, self.repository.repository, self.image_hash),
        ):
            result.add((Repository(namespace, repository), image_hash))
        return list(result)

    def set_provenance(self, provenance_data: List[ProvenanceLine]) -> None:
        """
        Sets the image's provenance. Internal function called by the Splitfile interpreter, shouldn't
        be called directly as it changes the image after it's been created.

        :param provenance_data: List of parsed Splitfile commands and their data.
        """
        self.engine.run_sql(
            SQL(
                """UPDATE {}.images SET provenance_data = %s WHERE
                            namespace = %s AND repository = %s AND image_hash = %s"""
            ).format(Identifier(SPLITGRAPH_META_SCHEMA)),
            (
                Json(provenance_data),
                self.repository.namespace,
                self.repository.repository,
                self.image_hash,
            ),
        )


def reconstruct_splitfile(
    provenance_data: List[ProvenanceLine],
    ignore_irreproducible: bool = False,
    source_replacement: Optional[Dict["Repository", str]] = None,
) -> List[str]:
    """
    Recreate the Splitfile that can be used to reconstruct an image.
    """

    if source_replacement is None:
        source_replacement = {}
    splitfile_commands = []
    for provenance_line in provenance_data:
        prov_type = provenance_line["type"]
        assert isinstance(prov_type, str)
        if prov_type in ("IMPORT", "SQL", "FROM"):
            splitfile_commands.append(
                _prov_command_to_splitfile(provenance_line, source_replacement)
            )
        elif prov_type in ("MOUNT", "CUSTOM"):
            if not ignore_irreproducible:
                raise SplitGraphError(
                    "Image used a Splitfile command %s" " that can't be reproduced!" % prov_type
                )
            splitfile_commands.append("# Irreproducible Splitfile command of type %s" % prov_type)
    return splitfile_commands


def _prov_command_to_splitfile(
    prov_data: ProvenanceLine,
    source_replacement: Dict["Repository", str],
) -> str:
    """
    Converts the image's provenance data stored by the Splitfile executor back to a Splitfile used to
    reconstruct it.

    :param prov_data: Provenance line for one command
    :param source_replacement: Replace repository imports with different versions
    :return: String with the Splitfile command.
    """
    from splitgraph.core.repository import Repository

    prov_type = prov_data["type"]
    assert isinstance(prov_type, str)

    if prov_type == "IMPORT":
        repo, image = (
            Repository(cast(str, prov_data["source_namespace"]), cast(str, prov_data["source"])),
            cast(str, prov_data["source_hash"]),
        )
        result = "FROM %s:%s IMPORT " % (str(repo), source_replacement.get(repo, image))
        result += ", ".join(
            "%s AS %s" % (tn if not q else "{" + tn.replace("}", "\\}") + "}", ta)
            for tn, ta, q in zip(
                cast(List[str], prov_data["tables"]),
                cast(List[str], prov_data["table_aliases"]),
                cast(List[bool], prov_data["table_queries"]),
            )
        )
        return result
    if prov_type == "FROM":
        repo = Repository(cast(str, prov_data["source_namespace"]), cast(str, prov_data["source"]))
        return "FROM %s:%s" % (str(repo), source_replacement.get(repo, prov_data["source_hash"]))
    if prov_type == "SQL":
        # Use the SQL validator/replacer to rewrite old image hashes into new hashes/tags.

        def image_mapper(repository: Repository, image_hash: str):
            new_image = (
                repository.to_schema() + ":" + source_replacement.get(repository, image_hash)
            )
            return new_image, new_image

        if source_replacement:
            _, replaced_sql = prepare_splitfile_sql(str(prov_data["sql"]), image_mapper)
        else:
            replaced_sql = str(prov_data["sql"])
        return "SQL " + "{" + replaced_sql.replace("}", "\\}") + "}"
    raise SplitGraphError("Cannot reconstruct provenance %s!" % prov_type)



"""Command line tools for building Splitgraph images from Singer taps, including using Splitgraph as a Singer target."""
import click

from splitgraph.commandline.common import ImageType
from splitgraph.hooks.data_source.base import prepare_new_image


@click.group(name="singer")
def singer_group():
    """Build Splitgraph images from Singer taps."""


@click.command(name="target")
@click.argument("image", type=ImageType(default="latest", repository_exists=False))
@click.option(
    "-d", "--delete-old", is_flag=True, help="Delete the old image at the end of ingestion"
)
@click.option(
    "-f",
    "--failure",
    type=click.Choice(["keep-both", "delete-old", "delete-new"]),
    help="What to do in case of a failure.",
    default="delete-new",
)
def singer_target(image, delete_old, failure):
    """
    Singer target that loads data into Splitgraph images.

    This will read data from the stdin from a Singer-compatible tap and load it into
    a Splitgraph image, merging data if the image already exists.

    Image must be of the format `[NAMESPACE/]REPOSITORY[:HASH_OR_TAG]` where `HASH_OR_TAG`
    is a tag of an existing image to base the image on. If the repository doesn't exist,
    it will be created.

    As this target consumes data from stdin, it will flush the records into a Splitgraph image. By
    default, it will only keep the image if the whole stream has been successfully consumed. To
    make this target completely follow the Singer spec (if it emits state, the records have been
    flushed), pass --failure=keep-both or --failure=delete-old. To delete the old image on success,
    pass --delete-old.
    """
    from splitgraph.ingestion.singer.db_sync import run_patched_sync

    repository, hash_or_tag = image
    base_image, new_image_hash = prepare_new_image(repository, hash_or_tag)
    run_patched_sync(repository, base_image, new_image_hash, delete_old, failure)


singer_group.add_command(singer_target)



import os
import sys

# "Export" an object of SystemConfigGetters, for getting config values that
# require operations with the system, like checking if files exist.
#
# Don't forget to add each method definition to the object after defining it.
from typing import List, Optional

from .argument_config import get_argument_config_value
from .environment_config import get_environment_config_value

SYSTEM_CONFIG_GETTERS = {}

VALID_CONFIG_FILE_NAMES = [".sgconfig", ".sgrc", ".splitgraph.config"]

HOME_SUB_DIR = ".splitgraph"


def is_file(filename: str) -> bool:
    return os.path.isfile(filename)


def file_exists(_dir: str, filename: str) -> bool:
    return is_file(os.path.join(_dir, filename))


def get_explicit_config_file_location() -> Optional[str]:
    """Get the explicitly defined location of config file, if defined.

    The location will either be defined in:

       * argument flag --config-file
       * or environment key SG_CONFIG_FILE

    In keeping with assumptions about priority, argument flag has
    higher priority than environment value.

    If the location is set, and points to an existing file, return location.

    Otherwise return None

    Print a warning if location is set but points to non-existing file.
    """

    key = "SG_CONFIG_FILE"

    explicit_location = get_argument_config_value(key, None) or get_environment_config_value(
        key, None
    )

    if explicit_location and is_file(explicit_location):
        return explicit_location
    if explicit_location and not is_file(explicit_location):
        sys.stderr.write("Warning: %s = %s is not a file" % (key, explicit_location))

    return None


def get_explicit_config_file_dirs() -> List[str]:
    """Get any explicitly defined config file directories,
    which are directories where we should search for files from
    VALID_CONFIG_FILE_NAMES.

    This list is defined similar to $PATH, as a colon (:) delimited
    string, either in:

        * argument flag --config-dirs
        * or environment key SG_CONFIG_DIRS

    Or a single directory defined in either

        * argument flag --config-dir
        * environment key SG_CONFIG_DIR

    If both plural and single are defined, join them together.

    Return a list of valid paths that are set, or an empty list.

    Print a warning if any paths to not exist.
    """

    plural_key = "SG_CONFIG_DIRS"
    single_key = "SG_CONFIG_DIR"

    explicit_plural_string = get_argument_config_value(
        plural_key, None
    ) or get_environment_config_value(plural_key, None)

    explicit_single_string = get_argument_config_value(
        single_key, None
    ) or get_environment_config_value(single_key, None)

    explicit_plural_list = explicit_plural_string.split(":") if explicit_plural_string else []
    explicit_single_list = [explicit_single_string] if explicit_single_string else []
    explicit_dir_list = explicit_plural_list + explicit_single_list
    unique_explicit_dir_list = list(set(explicit_dir_list))

    if len(unique_explicit_dir_list) != len(explicit_dir_list):
        num_duplicates = len(explicit_dir_list) - len(unique_explicit_dir_list)

        sys.stderr.write("Warning: %d duplicate SG_CONFIG_DIRS values found.\n" % num_duplicates)

    existing_unique_dir_list = [d for d in unique_explicit_dir_list if os.path.isdir(d)]

    if len(existing_unique_dir_list) != len(unique_explicit_dir_list):
        num_non_existing = len(unique_explicit_dir_list) - len(existing_unique_dir_list)

        sys.stderr.write(
            "Warning: %d non-existing SG_CONFIG_DIRS values found.\n" % num_non_existing
        )

    return existing_unique_dir_list


def get_config_file(default_return: None = None) -> Optional[str]:
    """
    Get the location of an existing SG_CONFIG_FILE on the system with
    a valid name. Do not attempt to parse the config file, just return
    its location.

    Otherwise, return default_return
    """
    explicit_location = get_explicit_config_file_location()

    if explicit_location:
        return explicit_location

    valid_dirs = [os.getcwd()]

    if os.environ.get("HOME", None) and os.path.isdir(
        os.path.join(os.environ["HOME"], HOME_SUB_DIR)
    ):
        valid_dirs.append(os.path.join(os.environ["HOME"], HOME_SUB_DIR))

    explicit_dirs = get_explicit_config_file_dirs()

    # Put explicit_dirs first to ensure higher priority given to explicit
    valid_dirs = explicit_dirs + valid_dirs

    matching_files = []
    for _dir in valid_dirs:
        matching_files.extend(
            [
                os.path.join(_dir, name)
                for name in VALID_CONFIG_FILE_NAMES
                if file_exists(_dir, name)
            ]
        )

    num_matching_files = len(matching_files)

    if num_matching_files > 0:
        first_matching_file = matching_files[0]

        if num_matching_files > 1:
            sys.stderr.write(
                "Warning: %d splitgraph config files found. \n Using %s \n"
                % (num_matching_files, first_matching_file)
            )

        return first_matching_file

    return default_return


# Don't forget to do this for each method you want in the object
SYSTEM_CONFIG_GETTERS["SG_CONFIG_FILE"] = get_config_file


def get_system_config_value(key: str, default_return: None = None) -> Optional[str]:
    return SYSTEM_CONFIG_GETTERS.get(key, lambda: default_return)()



from datetime import datetime
from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Set, cast

from psycopg2.extras import Json
from psycopg2.sql import SQL, Identifier

from splitgraph.config import SPLITGRAPH_API_SCHEMA
from splitgraph.core.engine import repository_exists
from splitgraph.core.image import IMAGE_COLS, Image
from splitgraph.core.output import truncate_line
from splitgraph.core.sql import select
from splitgraph.core.types import ProvenanceLine
from splitgraph.engine import ResultShape
from splitgraph.exceptions import ImageNotFoundError, RepositoryNotFoundError

if TYPE_CHECKING:
    from splitgraph.core.repository import Repository

_MAX_COMMENT_LEN = 4096


class ImageManager:
    """Collects various image-related functions."""

    def __init__(self, repository: "Repository") -> None:
        self.repository = repository
        self.engine = repository.engine

    def __call__(self) -> List[Image]:
        """Get all Image objects in the repository, ordered by their creation time (earliest first)."""
        result = []
        for image in self.engine.run_sql(
            select(
                "get_images",
                ",".join(IMAGE_COLS),
                schema=SPLITGRAPH_API_SCHEMA,
                table_args="(%s, %s)",
            ),
            (self.repository.namespace, self.repository.repository),
        ):
            result.append(self._make_image(image))
        return result

    def _make_image(self, img_tuple: Any) -> Image:
        r_dict = {k: v for k, v in zip(IMAGE_COLS, img_tuple)}
        r_dict.update(repository=self.repository)
        return Image(**r_dict)

    def by_tag(self, tag: str, raise_on_none: bool = True) -> Optional[Image]:
        """
        Returns an image with a given tag

        :param tag: Tag. 'latest' is a special case: it returns the most recent image in the repository.
        :param raise_on_none: Whether to raise an error or return None if the tag doesn't exist.
        """
        engine = self.engine
        if not repository_exists(self.repository):
            raise RepositoryNotFoundError("Unknown repository %s!" % str(self.repository))

        if tag == "latest":
            # Special case, return the latest commit from the repository.
            result = self.engine.run_sql(
                select(
                    "get_images",
                    ",".join(IMAGE_COLS),
                    schema=SPLITGRAPH_API_SCHEMA,
                    table_args="(%s,%s)",
                )
                + SQL(" ORDER BY created DESC LIMIT 1"),
                (self.repository.namespace, self.repository.repository),
                return_shape=ResultShape.ONE_MANY,
            )
            if result is None:
                raise ImageNotFoundError("No images found in %s!" % self.repository.to_schema())
            return self._make_image(result)

        result = engine.run_sql(
            select(
                "get_tagged_images",
                "image_hash",
                "tag = %s",
                schema=SPLITGRAPH_API_SCHEMA,
                table_args="(%s,%s)",
            ),
            (self.repository.namespace, self.repository.repository, tag),
            return_shape=ResultShape.ONE_ONE,
        )
        if result is None:
            if raise_on_none:
                schema = self.repository.to_schema()
                if tag == "HEAD":
                    raise ImageNotFoundError(
                        'No current checked out revision found for %s. Check one out with "sgr '
                        'checkout %s:image_hash".' % (schema, schema)
                    )
                raise ImageNotFoundError("Tag %s not found in repository %s" % (tag, schema))
            return None
        return self.by_hash(result)

    def by_hash(self, image_hash: str) -> Image:
        """
        Returns an image corresponding to a given (possibly shortened) image hash. If the image hash
        is ambiguous, raises an error. If the image does not exist, raises an error or returns None.

        :param image_hash: Image hash (can be shortened).
        :return: Image
        """
        result = self.engine.run_sql(
            select(
                "get_image",
                ",".join(IMAGE_COLS),
                schema=SPLITGRAPH_API_SCHEMA,
                table_args="(%s, %s, %s)",
            ),
            (self.repository.namespace, self.repository.repository, image_hash.lower()),
            return_shape=ResultShape.MANY_MANY,
        )
        if not result:
            raise ImageNotFoundError("No images starting with %s found!" % image_hash)
        if len(result) > 1:
            result = "Multiple suitable candidates found: \n * " + "\n * ".join(
                [r[0] for r in result]
            )
            raise ImageNotFoundError(result)
        return self._make_image(result[0])

    def __getitem__(self, key: str) -> Image:
        """Resolve an Image object from its tag or hash."""
        # Things we can have here: full hash, shortened hash or tag.
        # Users can always use by_hash or by_tag to be explicit -- this is just a shorthand. There's little
        # chance for ambiguity (why would someone have a hexadecimal tag that can be confused with a hash?)
        # so we can detect what the user meant in the future.
        # For HEAD (currently checked-out) images, raise if there's nothing
        # checked out (otherwise we'll fallback to hash and get an even more confusing message).
        return self.by_tag(key, raise_on_none=key == "HEAD") or self.by_hash(key)

    def get_all_child_images(self, start_image: str) -> Set[str]:
        """Get all children of `start_image` of any degree."""
        all_images = self()
        result_size = 1
        result = {start_image}
        while True:
            # Keep expanding the set of children until it stops growing
            for image in all_images:
                if image.parent_id in result:
                    result.add(image.image_hash)
            if len(result) == result_size:
                return result
            result_size = len(result)

    def get_all_parent_images(self, start_images: Set[str]) -> Set[str]:
        """Get all parents of the 'start_images' set of any degree."""
        parent = {image.image_hash: image.parent_id for image in self()}
        result = set(start_images)
        result_size = len(result)
        while True:
            # Keep expanding the set of parents until it stops growing
            result.update(
                {cast(str, parent[image]) for image in result if parent[image] is not None}
            )
            if len(result) == result_size:
                return result
            result_size = len(result)

    def add(
        self,
        parent_id: Optional[str],
        image: str,
        created: Optional[datetime] = None,
        comment: Optional[str] = None,
        provenance_data: Optional[List[ProvenanceLine]] = None,
    ) -> None:
        """
        Registers a new image in the Splitgraph image tree.

        Internal method used by actual image creation routines (committing, importing or pulling).

        :param parent_id: Parent of the image
        :param image: Image hash
        :param created: Creation time (defaults to current timestamp)
        :param comment: Comment (defaults to empty)
        :param provenance_data: Provenance data that can be used to reconstruct the image.
        """

        if comment:
            comment = truncate_line(comment, _MAX_COMMENT_LEN)

        self.engine.run_sql(
            SQL("SELECT {}.add_image(%s, %s, %s, %s, %s, %s, %s)").format(
                Identifier(SPLITGRAPH_API_SCHEMA)
            ),
            (
                self.repository.namespace,
                self.repository.repository,
                image,
                parent_id,
                created or datetime.utcnow(),
                comment,
                Json(provenance_data),
            ),
        )

    def add_batch(self, images: List[Image]) -> None:
        """
        Like add, but registers multiple images at the same time. Used in push/pull
        to avoid a roundtrip to the registry for each image
        :param images: List of Image objects. Namespace and repository will be patched
            with this repository.
        """
        now = datetime.utcnow()
        self.engine.run_sql_batch(
            SQL("SELECT {}.add_image(%s, %s, %s, %s, %s, %s, %s)").format(
                Identifier(SPLITGRAPH_API_SCHEMA)
            ),
            [
                (
                    self.repository.namespace,
                    self.repository.repository,
                    image.image_hash,
                    image.parent_id,
                    image.created or now,
                    image.comment,
                    Json(image.provenance_data),
                )
                for image in images
            ],
        )

    def delete(self, images: Iterable[str]) -> None:
        """
        Deletes a set of Splitgraph images from the repository. Note this doesn't check whether
        this will orphan some other images in the repository and can make the state of the repository
        invalid.

        Image deletions won't be replicated on push/pull (those can only add new images).

        :param images: List of image IDs
        """
        if not images:
            return
        # Maybe better to have ON DELETE CASCADE on the FK constraints instead of going through
        # all tables to clean up -- but then we won't get alerted when we accidentally try
        # to delete something that does have FKs relying on it.
        self.engine.run_sql_batch(
            SQL("SELECT {}.delete_image(%s, %s, %s)").format(Identifier(SPLITGRAPH_API_SCHEMA)),
            arguments=[(self.repository.namespace, self.repository.repository, i) for i in images],
        )

    def __iter__(self):
        return iter(self())



import json
import logging
import re
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Tuple, cast

import psycopg2
from psycopg2.sql import SQL, Composed, Identifier

from splitgraph.core.types import (
    Credentials,
    IntrospectionResult,
    MountError,
    PreviewResult,
    TableColumn,
    TableInfo,
    TableParams,
    TableSchema,
    dict_to_table_schema_params,
)
from splitgraph.engine import validate_type
from splitgraph.exceptions import DataSourceError, get_exception_name
from splitgraph.hooks.data_source.base import LoadableDataSource, MountableDataSource

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PsycopgEngine


class ForeignDataWrapperDataSource(MountableDataSource, LoadableDataSource, ABC):
    credentials_schema: Dict[str, Any] = {
        "type": "object",
    }

    params_schema: Dict[str, Any] = {
        "type": "object",
    }

    table_params_schema: Dict[str, Any] = {
        "type": "object",
    }

    commandline_help: str = ""
    commandline_kwargs_help: str = ""

    supports_mount = True
    supports_load = True

    @classmethod
    def from_commandline(cls, engine, commandline_kwargs) -> "ForeignDataWrapperDataSource":
        """Instantiate an FDW data source from commandline arguments."""
        from splitgraph.cloud.models import ExternalTableRequest

        # Normally these are supposed to be more user-friendly and FDW-specific (e.g.
        # not forcing the user to pass in a JSON with five levels of nesting etc).
        params = deepcopy(commandline_kwargs)

        # By convention, the "tables" object can be:
        #   * A list of tables to import
        #   * A dictionary table_name -> {"schema": schema, "options": table options}
        tables = params.pop("tables", None)
        if isinstance(tables, dict):
            tables = dict_to_table_schema_params(
                {k: ExternalTableRequest.parse_obj(v) for k, v in tables.items()}
            )

        # Extract credentials from the cmdline params
        credentials = Credentials({})
        for k in cast(Dict[str, Any], cls.credentials_schema.get("properties", {})).keys():
            if k in params:
                credentials[k] = params[k]

        result = cls(engine, credentials, params, tables)

        return result

    def get_user_options(self) -> Mapping[str, str]:
        return {}

    def get_server_options(self) -> Mapping[str, str]:
        return {}

    def get_table_options(
        self, table_name: str, tables: Optional[TableInfo] = None
    ) -> Dict[str, str]:
        tables = tables or self.tables

        if not isinstance(tables, dict):
            return {}

        return {
            k: str(v)
            for k, v in tables.get(table_name, cast(Tuple[TableSchema, TableParams], ({}, {})))[
                1
            ].items()
        }

    def get_table_schema(self, table_name: str, table_schema: TableSchema) -> TableSchema:
        # Hook to override the table schema that's passed in
        return table_schema

    def get_remote_schema_name(self) -> str:
        """Override this if the FDW supports IMPORT FOREIGN SCHEMA"""
        raise NotImplementedError

    @abstractmethod
    def get_fdw_name(self):
        pass

    def mount(
        self,
        schema: str,
        tables: Optional[TableInfo] = None,
        overwrite: bool = True,
    ) -> Optional[List[MountError]]:
        self._validate_table_params(tables)
        tables = tables or self.tables or []

        fdw = self.get_fdw_name()
        server_id = "%s_%s_server" % (schema, fdw)
        init_fdw(
            self.engine,
            server_id,
            self.get_fdw_name(),
            self.get_server_options(),
            self.get_user_options(),
            overwrite=overwrite,
        )

        # Allow mounting tables into existing schemas
        self.engine.run_sql(SQL("CREATE SCHEMA IF NOT EXISTS {}").format(Identifier(schema)))

        errors = self._create_foreign_tables(schema, server_id, tables)
        if errors:
            for e in errors:
                logging.warning("Error mounting %s: %s: %s", e.table_name, e.error, e.error_text)
        return errors

    def _create_foreign_tables(
        self, schema: str, server_id: str, tables: TableInfo
    ) -> List[MountError]:
        if isinstance(tables, list):
            try:
                remote_schema = self.get_remote_schema_name()
            except NotImplementedError:
                raise NotImplementedError(
                    "The FDW does not support IMPORT FOREIGN SCHEMA! Pass a tables dictionary."
                )
            return import_foreign_schema(self.engine, schema, remote_schema, server_id, tables)
        else:
            for table_name, (table_schema, _) in tables.items():
                logging.info("Mounting table %s", table_name)
                query, args = create_foreign_table(
                    schema,
                    server_id,
                    table_name,
                    schema_spec=self.get_table_schema(table_name, table_schema),
                    extra_options=self.get_table_options(table_name, tables),
                )
                self.engine.run_sql(query, args)
            return []

    def introspect(self) -> IntrospectionResult:
        # Ability to override introspection by e.g. contacting the remote database without having
        # to mount the actual table. By default, just call out into mount()

        # Local import here since this data source gets imported by the commandline entry point
        import jsonschema

        from splitgraph.core.common import get_temporary_table_id

        tmp_schema = get_temporary_table_id()
        try:
            mount_errors = self.mount(tmp_schema) or []

            table_options = dict(self._get_foreign_table_options(tmp_schema))

            # Sanity check for adapters: validate the foreign table options that we get back
            # to make sure they're still appropriate.
            for v in table_options.values():
                jsonschema.validate(v, self.table_params_schema)

            result = IntrospectionResult(
                {
                    t: (
                        self.engine.get_full_table_schema(tmp_schema, t),
                        cast(TableParams, table_options.get(t, {})),
                    )
                    for t in self.engine.get_all_tables(tmp_schema)
                }
            )

            # Add errors to the result as well
            for error in mount_errors:
                result[error.table_name] = error
            return result
        finally:
            self.engine.rollback()
            self.engine.delete_schema(tmp_schema)
            self.engine.commit()

    def _preview_table(self, schema: str, table: str, limit: int = 10) -> List[Dict[str, Any]]:
        from splitgraph.engine import ResultShape

        result_json = cast(
            List[Dict[str, Any]],
            self.engine.run_sql(
                SQL("SELECT row_to_json(t.*) FROM {}.{} t LIMIT %s").format(
                    Identifier(schema), Identifier(table)
                ),
                (limit,),
                return_shape=ResultShape.MANY_ONE,
            ),
        )
        return result_json

    def _get_foreign_table_options(self, schema: str) -> List[Tuple[str, Dict[str, Any]]]:
        """
        Get a list of options the foreign tables in this schema were instantiated with
        :return: List of tables and their options
        """
        # We use this to suggest table options during introspection. With FDWs, we do this by
        # mounting the data source first (using IMPORT FOREIGN SCHEMA) and then scraping the
        # foreign table options it inferred.

        # Downstream FDWs can override this: if they remap some table options into different
        # FDW options (e.g. "remote_schema" on the data source side turns into "schema" on the
        # FDW side), they have to map them back in this routine (otherwise the introspection will
        # suggest "schema", which is wrong.

        # This is also used for type remapping, since table params can only be strings on PG.
        # TODO: this will lead to a bunch of serialization/deserialization code duplication
        #   in data sources. One potential solution is, at least in Multicorn-backed wrappers
        #   that we control, passing a JSON as a single table param instead.

        return cast(
            List[Tuple[str, Dict[str, str]]],
            self.engine.run_sql(
                "SELECT foreign_table_name, json_object_agg(option_name, option_value) "
                "FROM information_schema.foreign_table_options "
                "WHERE foreign_table_schema = %s GROUP BY foreign_table_name",
                (schema,),
            ),
        )

    def preview(self, tables: Optional[TableInfo]) -> PreviewResult:
        # Preview data in tables mounted by this FDW / data source

        # Local import here since this data source gets imported by the commandline entry point
        from splitgraph.core.common import get_temporary_table_id

        tmp_schema = get_temporary_table_id()
        try:
            # Seed the result with the tables that failed to mount
            result = PreviewResult(
                {e.table_name: e for e in self.mount(tmp_schema, tables=tables) or []}
            )

            # Commit so that errors don't cancel the mount
            self.engine.commit()
            for t in self.engine.get_all_tables(tmp_schema):
                if t in result:
                    continue
                try:
                    result[t] = self._preview_table(tmp_schema, t)
                except psycopg2.DatabaseError as e:
                    logging.warning("Could not preview data for table %s", t, exc_info=e)
                    result[t] = MountError(
                        table_name=t, error=get_exception_name(e), error_text=str(e).strip()
                    )
            return result
        finally:
            self.engine.rollback()
            self.engine.delete_schema(tmp_schema)
            self.engine.commit()

    def _load(self, schema: str, tables: Optional[TableInfo] = None):
        from splitgraph.core.common import get_temporary_table_id

        tmp_schema = get_temporary_table_id()
        try:
            errors = self.mount(tmp_schema, tables=tables)
            if errors:
                raise DataSourceError("Error mounting tables for load")

            self.engine.commit()

            for t in self.engine.get_all_tables(tmp_schema):
                try:
                    self.engine.copy_table(tmp_schema, t, schema, t)
                    self.engine.commit()
                except psycopg2.DatabaseError as e:
                    logging.exception("Error ingesting table %s", t, exc_info=e)
                    raise DataSourceError(
                        "Error ingesting table %s: %s: %s" % (t, get_exception_name(e), e)
                    )
        finally:
            self.engine.rollback()
            self.engine.delete_schema(tmp_schema)
            self.engine.commit()


def init_fdw(
    engine: "PsycopgEngine",
    server_id: str,
    wrapper: str,
    server_options: Optional[Mapping[str, Optional[str]]] = None,
    user_options: Optional[Mapping[str, str]] = None,
    role: Optional[str] = None,
    overwrite: bool = True,
) -> None:
    """
    Sets up a foreign data server on the engine.

    :param engine: PostgresEngine
    :param server_id: Name to call the foreign server, must be unique. Will be deleted if exists.
    :param wrapper: Name of the foreign data wrapper (must be installed as an extension on the engine)
    :param server_options: Dictionary of FDW options
    :param user_options: Dictionary of user options
    :param role: The name of the role for which the user mapping is created; defaults to public.
    :param overwrite: If the server already exists, delete and recreate it.
    """
    from psycopg2.sql import SQL, Identifier

    if overwrite:
        engine.run_sql(SQL("DROP SERVER IF EXISTS {} CASCADE").format(Identifier(server_id)))

    create_server = SQL("CREATE SERVER IF NOT EXISTS {} FOREIGN DATA WRAPPER {}").format(
        Identifier(server_id), Identifier(wrapper)
    )

    if server_options:
        server_keys, server_vals = zip(*server_options.items())
        create_server += _format_options(server_keys)
        engine.run_sql(create_server, [str(v) for v in server_vals])
    else:
        engine.run_sql(create_server)

    if user_options:
        create_mapping = SQL("CREATE USER MAPPING IF NOT EXISTS FOR {} SERVER {}").format(
            SQL("PUBLIC") if role is None else Identifier(role), Identifier(server_id)
        )
        user_keys, user_vals = zip(*user_options.items())
        create_mapping += _format_options(user_keys)
        engine.run_sql(create_mapping, [str(v) for v in user_vals])


def _format_options(option_names):
    from psycopg2.sql import SQL, Identifier

    return (
        SQL(" OPTIONS (")
        + SQL(",").join(Identifier(o) + SQL(" %s") for o in option_names)
        + SQL(")")
    )


def import_foreign_schema(
    engine: "PsycopgEngine",
    mountpoint: str,
    remote_schema: str,
    server_id: str,
    tables: List[str],
    options: Optional[Dict[str, str]] = None,
) -> List[MountError]:
    from psycopg2.sql import SQL, Identifier

    # Construct a query: import schema limit to (%s, %s, ...) from server mountpoint_server into mountpoint
    query = SQL("IMPORT FOREIGN SCHEMA {} ").format(Identifier(remote_schema))
    if tables:
        query += SQL("LIMIT TO (") + SQL(",").join(Identifier(t) for t in tables) + SQL(")")
    query += SQL("FROM SERVER {} INTO {}").format(Identifier(server_id), Identifier(mountpoint))
    args: List[str] = []

    if options:
        query += _format_options(options.keys())
        args.extend(options.values())

    engine.run_sql(query, args)

    # Some of our FDWs use the PG notices as a side channel to pass errors and information back
    # to the user, as IMPORT FOREIGN SCHEMA is all-or-nothing. If some of the tables failed to
    # mount, we return those + the reasons.
    import_errors: List[MountError] = []
    if engine.notices:
        for notice in engine.notices:
            match = re.match(r".*SPLITGRAPH: (.*)\n", notice)
            if not match:
                continue
            notice_j = json.loads(match.group(1))
            import_errors.append(
                MountError(
                    table_name=notice_j["table_name"],
                    error=notice_j["error"],
                    error_text=notice_j["error_text"],
                )
            )

    return import_errors


def _identifier(c: str) -> Identifier:
    """
    Create a psycopg2 composable Identifier escaping percentage signs in column names. These
    cause issues when using argument interpolation, as psycopg2 treats them as arguments.
    """
    return Identifier(c.replace("%", "%%").replace("{", "{{").replace("}", "}}"))


def create_foreign_table(
    schema: str,
    server: str,
    table_name: str,
    schema_spec: TableSchema,
    extra_options: Optional[Dict[str, str]] = None,
) -> Tuple[Composed, List[str]]:
    table_options = extra_options or {}

    query = SQL("CREATE FOREIGN TABLE {}.{} (").format(Identifier(schema), Identifier(table_name))
    query += SQL(",".join("{} %s " % validate_type(col.pg_type) for col in schema_spec)).format(
        *(_identifier(col.name) for col in schema_spec)
    )
    query += SQL(") SERVER {}").format(Identifier(server))

    args: List[str] = []
    if table_options:
        table_opts, table_optvals = zip(*table_options.items())
        query += SQL(" OPTIONS(")
        query += SQL(",").join(_identifier(o) + SQL(" %s") for o in table_opts) + SQL(");")
        args.extend(table_optvals)

    for col in schema_spec:
        if col.comment:
            query += SQL("COMMENT ON COLUMN {}.{}.{} IS %s;").format(
                Identifier(schema),
                Identifier(table_name),
                _identifier(col.name),
            )
            args.append(col.comment)
    return query, args


class PostgreSQLDataSource(ForeignDataWrapperDataSource):
    @classmethod
    def get_name(cls) -> str:
        return "PostgreSQL"

    @classmethod
    def get_description(cls) -> str:
        return (
            "Data source for PostgreSQL databases that supports live querying, "
            "based on postgres_fdw"
        )

    credentials_schema = {
        "type": "object",
        "properties": {
            "username": {
                "type": "string",
                "title": "Username",
            },
            "password": {
                "type": "string",
                "title": "Password",
            },
        },
        "required": ["username", "password"],
    }

    params_schema = {
        "type": "object",
        "properties": {
            "host": {
                "type": "string",
                "title": "Host",
                "description": "Hostname or IP address of the PostgreSQL instance",
            },
            "port": {"type": "integer", "title": "Port", "description": "Port of the database"},
            "dbname": {"type": "string", "title": "Database", "description": "Database name"},
            "remote_schema": {
                "type": "string",
                "title": "Schema name",
                "description": "Schema name to import data from",
                "default": "public",
            },
        },
        "required": ["host", "port", "dbname", "remote_schema"],
    }

    table_params_schema = {"type": "object"}

    commandline_help: str = """Mount a Postgres database.

Mounts a schema on a remote Postgres database as a set of foreign tables locally."""

    commandline_kwargs_help: str = """dbname: Database name (required)
remote_schema: Remote schema name (required)
extra_server_args: Dictionary of extra arguments to pass to the foreign server
tables: Tables to mount (default all). If a list, then will use IMPORT FOREIGN SCHEMA.
If a dictionary, must have the format
    {"table_name": {"schema": {"col_1": "type_1", ...},
                    "options": {[get passed to CREATE FOREIGN TABLE]}}}.
    """

    _icon_file = "postgresql.svg"

    def get_server_options(self):
        return {
            "host": self.params["host"],
            "port": str(self.params["port"]),
            "dbname": self.params["dbname"],
            **self.params.get("extra_server_args", {}),
        }

    def get_user_options(self):
        return {"user": self.credentials["username"], "password": self.credentials["password"]}

    def get_table_options(self, table_name: str, tables: Optional[TableInfo] = None):
        options = super().get_table_options(table_name, tables)
        options["schema_name"] = self.params["remote_schema"]
        return options

    def get_fdw_name(self):
        return "postgres_fdw"

    def get_remote_schema_name(self) -> str:
        return str(self.params["remote_schema"])


class MongoDataSource(ForeignDataWrapperDataSource):
    credentials_schema = {
        "type": "object",
        "properties": {
            "username": {"type": "string", "title": "Username"},
            "password": {"type": "string", "title": "Password"},
        },
        "required": ["username", "password"],
    }

    params_schema = {
        "type": "object",
        "properties": {
            "host": {"type": "string", "title": "Host"},
            "port": {"type": "integer", "title": "Port"},
        },
        "required": ["host", "port"],
    }

    table_params_schema = {
        "type": "object",
        "properties": {
            "database": {"type": "string", "title": "Database name"},
            "collection": {"type": "string", "title": "Collection name"},
        },
        "required": ["database", "collection"],
    }

    commandline_help = """Mount a Mongo database.

Mounts one or more collections on a remote Mongo database as a set of foreign tables locally."""

    commandline_kwargs_help = """tables: A dictionary of form
```
{
    "table_name": {
        "schema": {"col1": "type1"...},
        "options": {"database": <dbname>, "collection": <collection>}
    }
}
```
"""

    _icon_file = "mongodb.svg"

    @classmethod
    def get_name(cls) -> str:
        return "MongoDB"

    @classmethod
    def get_description(cls) -> str:
        return "Data source for MongoDB databases that supports live querying, based on mongo_fdw"

    def get_server_options(self):
        return {
            "address": self.params["host"],
            "port": str(self.params["port"]),
        }

    def get_user_options(self):
        return {"username": self.credentials["username"], "password": self.credentials["password"]}

    def get_fdw_name(self):
        return "mongo_fdw"

    def get_table_schema(self, table_name, table_schema):
        # Add the "_id" column to the schema if it's not already there.
        if any(c.name == "_id" for c in table_schema):
            return table_schema

        return table_schema + [TableColumn(table_schema[-1].ordinal + 1, "_id", "NAME", False)]


class MySQLDataSource(ForeignDataWrapperDataSource):
    credentials_schema = {
        "type": "object",
        "properties": {
            "username": {"type": "string", "title": "Username"},
            "password": {"type": "string", "title": "Password"},
        },
        "required": ["username", "password"],
    }

    params_schema = {
        "type": "object",
        "properties": {
            "host": {"type": "string", "title": "Host"},
            "port": {"type": "integer", "title": "Port"},
            "dbname": {"type": "string", "title": "Database name"},
        },
        "required": ["host", "port", "dbname"],
    }

    commandline_help: str = """Mount a MySQL database.

Mounts a schema on a remote MySQL database as a set of foreign tables locally."""

    commandline_kwargs_help: str = """dbname: Remote MySQL database name (required)
tables: Tables to mount (default all). If a list, then will use IMPORT FOREIGN SCHEMA.
If a dictionary, must have the format
    {"table_name": {"schema": {"col_1": "type_1", ...},
                    "options": {[get passed to CREATE FOREIGN TABLE]}}}.
        """

    _icon_file = "mysql.svg"

    @classmethod
    def get_name(cls) -> str:
        return "MySQL"

    @classmethod
    def get_description(cls) -> str:
        return "Data source for MySQL databases that supports live querying, based on mysql_fdw"

    def get_server_options(self):
        return {
            "host": self.params["host"],
            "port": str(self.params["port"]),
            **self.params.get("extra_server_args", {}),
        }

    def get_user_options(self):
        return {"username": self.credentials["username"], "password": self.credentials["password"]}

    def get_table_options(self, table_name: str, tables: Optional[TableInfo] = None):
        options = super().get_table_options(table_name, tables)
        options["dbname"] = options.get("dbname", self.params["dbname"])
        return options

    def get_fdw_name(self):
        return "mysql_fdw"

    def get_remote_schema_name(self) -> str:
        return str(self.params["dbname"])


class ElasticSearchDataSource(ForeignDataWrapperDataSource):
    credentials_schema = {
        "type": "object",
        "properties": {
            "username": {"type": "string", "title": "Username"},
            "password": {"type": "string", "title": "Password"},
        },
    }

    params_schema = {
        "type": "object",
        "properties": {
            "host": {"type": "string", "title": "Host"},
            "port": {"type": "integer", "title": "Port"},
        },
        "required": ["host", "port"],
    }

    table_params_schema = {
        "type": "object",
        "properties": {
            "index": {
                "type": "string",
                "title": "Index name or pattern",
                "description": 'ES index name or pattern to use, for example, "events-*"',
            },
            "type": {
                "type": "string",
                "title": "doc_type (ES6 or earlier)",
                "description": "Pre-ES7 doc_type, not required in ES7 or later",
            },
            "query_column": {
                "type": "string",
                "title": "Query column name",
                "description": "Allows you to do `WHERE query = '{" "match" ": ...}'`",
                "default": "query",
            },
            "score_column": {
                "type": "string",
                "title": "Score column name",
                "description": "Name of the column with the document score",
            },
            "scroll_size": {
                "type": "integer",
                "title": "Scroll size",
                "description": "Fetch size, default 1000",
                "default": 1000,
            },
            "scroll_duration": {
                "type": "string",
                "title": "Scroll duration",
                "description": "How long to hold the scroll context open for, default 10m",
                "default": "10m",
            },
        },
        "required": ["index"],
    }

    commandline_help = """Mount an ElasticSearch instance.

Mount a set of tables proxying to a remote ElasticSearch index.

This uses a fork of postgres-elasticsearch-fdw behind the scenes. You can add a column
`query` to your table and set it as `query_column` to pass advanced ES queries and aggregations.
For example:

\b
```
$ sgr mount elasticsearch target_schema -c elasticsearch:9200 -o@- <<EOF
    {
      "tables": {
        "table_1": {
          "schema": {
            "id": "text",
            "@timestamp": "timestamp",
            "query": "text",
            "col_1": "text",
            "col_2": "boolean"
          },
          "options": {
              "index": "index-pattern*",
              "rowid_column": "id",
              "query_column": "query"
          }
        }
      }
    }
EOF
\b
```
"""

    _icon_file = "elasticsearch.svg"

    @classmethod
    def get_name(cls) -> str:
        return "Elasticsearch"

    @classmethod
    def get_description(cls) -> str:
        return (
            "Data source for Elasticsearch indexes that supports live querying, "
            "based on pg_es_fdw"
        )

    def get_server_options(self):
        result = {
            "wrapper": "pg_es_fdw.ElasticsearchFDW",
            "host": self.params["host"],
            "port": self.params["port"],
        }

        for key in ["username", "password"]:
            if key in self.credentials:
                result[key] = self.credentials[key]

        return result

    def get_fdw_name(self):
        return "multicorn"



import logging
import os
from pathlib import Path

from splitgraph.config.config import patch_config
from splitgraph.config.export import overwrite_config
from splitgraph.config.system_config import HOME_SUB_DIR


def patch_and_save_config(config, patch):

    config_path = config["SG_CONFIG_FILE"]
    if not config_path:
        # Default to creating a config in the user's homedir rather than local.
        config_dir = Path(os.environ["HOME"]) / Path(HOME_SUB_DIR)
        config_path = config_dir / Path(".sgconfig")
        logging.debug("No config file detected, creating one at %s" % config_path)
        config_dir.mkdir(exist_ok=True, parents=True)
    else:
        logging.debug("Updating the existing config file at %s" % config_path)
    new_config = patch_config(config, patch)
    overwrite_config(new_config, config_path)
    return str(config_path)
